from __future__ import annotations

import json
import logging
import os
import warnings

import anndata
import joblib
import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse as scp
import torch
from pynndescent import NNDescent
from scanpy._utils import check_nonnegative_integers

from popv import _utils


class Process_Query:
    """
    Processes the query and reference dataset in preparation for the annotation pipeline.

    Parameters
    ----------
    query_adata
        AnnData of query cells
    ref_adata
        AnnData of reference cells. Can contain only latent spaces if prediction_mode is not 'retrain'.
    ref_labels_key
        Key in obs field of reference AnnData with cell-type information
    ref_batch_key
        List of Keys (or None) in obs field of reference AnnData to
        use as batch covariate
    cl_obo_folder
        Folder containing the cell-type obo for OnClass, ontologies for OnClass and nlp embedding of cell-types.
        Passing a list will use element 1 as obo, element 2 as ontologies and element 3 as nlp embedding.
        Setting it to false will disable ontology use.
    query_batch_key
        Key in obs field of query adata for batch information.
    query_layer_key
        If not None, expects raw_count data in query_layer_key.
    ref_layer_key
        If not None, expects raw_count data in ref_layer_key.
    prediction_mode
        Execution mode of cell-type annotation.
        "retrain": Train all prediction models and saves them to disk if save_path_trained_models is not None.
        "inference": Classify all cells based on pretrained models.
        "fast": Fast inference using only query cells and single epoch in scArches.
    unknown_celltype_label
        Label for cells without a known cell-type.
    n_samples_per_label
        Reference AnnData will be subset to these amount of cells per cell-type to increase speed.
    pretrained_scvi_path
        If path is None, will train scVI from scratch. Else if
        pretrained_path is set and all the genes in the pretrained models are present
        in query adata, will train the scARCHES version of scVI and scANVI, resulting in
        faster training times.
    relabel_reference_cells
        If True, will relabel reference cells with cell-type information from query cells in inference mode.
    save_path_trained_models
        If mode=='retrain' saves models to this directory. Otherwise trained models are expected in this folder.
    hvg
        If Int, subsets data to n highly variable genes according to `sc.pp.highly_variable_genes`
    """

    def __init__(
        self,
        query_adata: anndata.AnnData,
        ref_adata: anndata.AnnData | bool,
        ref_labels_key: str,
        ref_batch_key: str,
        cl_obo_folder: list | str | bool,
        query_batch_key: str | None = None,
        query_layer_key: str | None = None,
        ref_layer_key: str | None = None,
        prediction_mode: str | None = "retrain",
        unknown_celltype_label: str = "unknown",
        n_samples_per_label: int | None = 300,
        save_path_trained_models: str = "tmp/",
        pretrained_scvi_path: str | None = None,
        relabel_reference_cells: bool = False,
        hvg: int | None = 4000,
    ) -> None:
        if ref_adata.X.sum() == 0:  # Minified object
            if prediction_mode == "retrain":
                ValueError("Reference dataset needs to contain gene expression to retrain models.")
            self.process_reference = False
        else:
            self.process_reference = True
            del ref_adata.obsm
            del ref_adata.uns
        del query_adata.obsm
        del query_adata.uns
        del query_adata.raw
        del ref_adata.raw
        if relabel_reference_cells:
            del ref_adata.uns["prediction_keys"]

        if cl_obo_folder is False:
            self.cl_obo_file = False
            self.cl_ontology_file = False
            self.nlp_emb_file = False
        elif isinstance(cl_obo_folder, list):
            self.cl_obo_file = cl_obo_folder[0]
            self.cl_ontology_file = cl_obo_folder[1]
            self.nlp_emb_file = cl_obo_folder[2]
        else:
            self.cl_obo_file = os.path.join(cl_obo_folder, "cl_popv.json")
            self.cl_ontology_file = os.path.join(cl_obo_folder, "cl.ontology")
            self.nlp_emb_file = os.path.join(cl_obo_folder, "cl.ontology.nlp.emb")
        if self.cl_obo_file:
            try:
                with open(self.cl_obo_file):
                    pass
            except FileNotFoundError as err:
                raise FileNotFoundError(f"{self.cl_obo_file} doesn't exist. Check that folder exists.") from err
        self.setup_dict = {
            "ref_labels_key": ref_labels_key,
            "ref_batch_key": ref_batch_key,
            "unknown_celltype_label": unknown_celltype_label,
        }
        self.ref_labels_key = ref_labels_key
        self.unknown_celltype_label = unknown_celltype_label
        self.n_samples_per_label = n_samples_per_label
        self.batch_key = {"reference": ref_batch_key, "query": query_batch_key}

        os.makedirs(save_path_trained_models, exist_ok=True)
        self.save_path_trained_models = save_path_trained_models
        self.pretrained_scvi_path = pretrained_scvi_path

        self.prediction_mode = prediction_mode
        json_path = os.path.join(save_path_trained_models, "preprocessing.json")
        if prediction_mode != "retrain" and not os.path.exists(json_path):
            raise ValueError(f"Configuration {json_path} doesn't exist. Set mode='retrain' to reprocess.")
        if prediction_mode != "retrain":
            with open(json_path) as f:
                data = json.load(f)
            from scvi.model.base._archesmixin import _pad_and_sort_query_anndata

            if not set(data["gene_names"]).issubset(set(query_adata.var_names)):
                _pad_and_sort_query_anndata(
                    query_adata,
                    reference_var_names=pd.Index(data["gene_names"]),
                    inplace=True,
                )
            self.label_categories = data["label_categories"]
            self.genes = data["gene_names"]
        else:
            self.label_categories = None
            self.genes = None

        if self.pretrained_scvi_path or self.prediction_mode != "retrain":
            if self.pretrained_scvi_path is None:
                self.pretrained_scvi_path = os.path.join(self.save_path_trained_models, "scvi")
            pretrained_scvi_genes = torch.load(
                os.path.join(self.pretrained_scvi_path, "model.pt"),
                map_location="cpu",
                weights_only=False,
            )["var_names"]
            if self.genes is not None and not np.array_equal(pretrained_scvi_genes, self.genes):
                warnings.warn(
                    "Pretrained scVI model and query dataset contain different genes. Retrain models or disable scVI.",
                    UserWarning,
                    stacklevel=2,
                )
            elif self.genes is None:
                self.genes = list(pretrained_scvi_genes)
        scanvi_path = os.path.join(self.save_path_trained_models, "scanvi/model.pt")
        if os.path.exists(scanvi_path) and self.prediction_mode != "retrain":
            pretrained_scanvi_genes = torch.load(
                scanvi_path,
                map_location="cpu",
                weights_only=False,
            )["var_names"]
            if not np.array_equal(pretrained_scanvi_genes, self.genes):
                warnings.warn(
                    "Pretrained scANVI model and query dataset contain different genes. Retrain models or disable scANVI.",
                    UserWarning,
                    stacklevel=2,
                )
        onclass_path = os.path.join(self.save_path_trained_models, "OnClass.npz")
        if os.path.exists(onclass_path) and prediction_mode != "retrain":
            onclass_model = np.load(
                onclass_path,
                allow_pickle=True,
            )
            if not np.array_equal(onclass_model["genes"], self.genes):
                warnings.warn(
                    "Pretrained scANVI model and query dataset contain different genes. Retrain models or disable scANVI.",
                    UserWarning,
                    stacklevel=2,
                )
        os.makedirs(self.save_path_trained_models, exist_ok=True)

        if self.genes is not None:
            if not set(self.genes).issubset(set(query_adata.var_names)):
                raise ValueError(
                    "Query dataset misses genes that were used for reference model training. Retrain reference model, set mode='retrain'"
                )
            if hvg is not None:
                raise ValueError("Highly variable gene selection is not available if using trained reference model.")
        else:
            gene_intersection = np.intersect1d(ref_adata.var_names, query_adata.var_names)
            if hvg is not None and len(gene_intersection) > hvg:
                expressed_genes, _ = sc.pp.filter_genes(ref_adata[:, gene_intersection], min_cells=200, inplace=False)
                subset_genes = gene_intersection[expressed_genes]
                if len(subset_genes) > hvg:
                    highly_variable_genes = sc.pp.highly_variable_genes(
                        ref_adata[:, subset_genes].copy(),
                        n_top_genes=hvg,
                        subset=False,
                        flavor="seurat_v3",
                        inplace=False,
                        # batch_key=ref_batch_key,
                        span=1.0,
                    )["highly_variable"]
                    self.genes = list(ref_adata[:, subset_genes].var_names[highly_variable_genes])
                else:
                    self.genes = list(subset_genes)
            else:
                self.genes = list(gene_intersection)

        if self.process_reference:
            self.ref_adata = self._setup_dataset(
                ref_adata,
                key="reference",
                layer_key=ref_layer_key,
                add_meta="_ref",
            )
            self._check_validity_anndata(self.ref_adata, "reference")
        else:
            self.ref_adata = ref_adata
        self.query_adata = self._setup_dataset(
            query_adata,
            key="query",
            add_meta="_query",
            layer_key=query_layer_key,
        )
        self._check_validity_anndata(self.query_adata, "query")

        self._preprocess()

    def _check_validity_anndata(self, adata, key):
        if not check_nonnegative_integers(adata.layers["scvi_counts"]):
            raise ValueError(f"Make sure input {key} adata contains raw_counts")
        if not len(set(adata.var_names)) == len(adata.var_names):
            raise ValueError(f"{key} dataset contains multiple genes with same gene name.")
        if adata.n_obs == 0:
            raise ValueError(f"{key} anndata has no cells.")
        if adata.n_vars == 0:
            raise ValueError(f"{key} anndata has no genes.")

    def _setup_dataset(self, adata, key, add_meta="", layer_key=None):
        if isinstance(self.batch_key[key], list):
            adata.obs["_batch_annotation"] = adata.obs[self.batch_key[key]].astype(str).sum(1).astype("category")
        elif isinstance(self.batch_key[key], str):
            adata.obs["_batch_annotation"] = adata.obs[self.batch_key[key]]
        else:
            adata.obs["_batch_annotation"] = self.unknown_celltype_label

        if layer_key is not None:
            adata.X = scp.csr_matrix(adata.layers[layer_key])
        del adata.layers
        adata = adata[:, self.genes].copy()

        zero_cell_names = adata[np.array(adata.X.sum(1) < 30).flatten()].obs_names
        adata.uns["Filtered_cells"] = list(zero_cell_names)
        sc.pp.filter_cells(adata, min_counts=30, inplace=True)
        if len(zero_cell_names) > 0:
            logging.warning(
                f"The following cells will be excluded from annotation because they have low expression:{zero_cell_names}."
            )

        adata.obs["_batch_annotation"] = adata.obs["_batch_annotation"].astype(str) + add_meta
        adata.obs["_dataset"] = key
        adata.obs["_batch_annotation"] = adata.obs["_batch_annotation"].astype("category")
        adata.layers["scvi_counts"] = adata.X.copy()
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)

        adata.obs["_reference_labels_annotation"] = adata.obs.get(self.ref_labels_key, None)

        # subsample the reference cells used for training certain models
        if key == "reference":
            adata.obs["_labels_annotation"] = adata.obs[self.ref_labels_key]
            if self.n_samples_per_label is not None:
                adata.obs["_ref_subsample"] = False
                subsample_idx = _utils.subsample_dataset(
                    adata,
                    self.ref_labels_key,
                    n_samples_per_label=self.n_samples_per_label,
                    ignore_label=[self.unknown_celltype_label],
                )
                adata.obs.loc[subsample_idx, "_ref_subsample"] = True
            else:
                adata.obs["_ref_subsample"] = True
            adata.layers["scaled"] = adata.X.copy()
            sc.pp.scale(adata, max_value=10, layer="scaled", zero_center=False)
            sc.pp.pca(adata, layer="scaled", zero_center=False)
            reference_features = adata.obsm["X_pca"]
            index = NNDescent(reference_features, n_neighbors=30, metric="euclidean")
            index_path = os.path.join(self.save_path_trained_models, "pynndescent_index.joblib")
            joblib.dump(index, open(index_path, "wb"))
        else:
            adata.obs["_labels_annotation"] = self.unknown_celltype_label
            adata.obs["_ref_subsample"] = False
            adata.layers["scaled"] = adata.X.copy()
            adata.layers["scaled"] /= self.ref_adata.var["std"].values
            adata.layers["scaled"].data = np.clip(adata.layers["scaled"].data, -10, 10)
            adata.layers["scaled"] = adata.layers["scaled"].tocsr()
            adata.obsm["X_pca"] = np.array(adata.layers["scaled"] @ self.ref_adata.varm["PCs"])
        return adata

    def _preprocess(self):
        if self.prediction_mode == "fast":
            self.adata = self.query_adata
        else:
            obsm_dtype = {key: value.dtype for key, value in self.ref_adata.obsm.items()}
            self.adata = anndata.concat(
                (self.ref_adata, self.query_adata),
                axis=0,
                label="_dataset",
                keys=["ref", "query"],
                join="outer",
                fill_value=self.unknown_celltype_label,
                merge="first",
                uns_merge="first",
            )
            self.adata.obsm = {
                key: pd.DataFrame(value).apply(pd.to_numeric, errors="coerce").astype(obsm_dtype[key]).to_numpy()
                for key, value in self.adata.obsm.items()
                if key in obsm_dtype
            }
        del self.query_adata, self.ref_adata
        self.adata.obs["_labels_annotation"] = self.adata.obs["_labels_annotation"].fillna(self.unknown_celltype_label)
        self.adata.obs["_labelled_train_indices"] = np.logical_and(
            self.adata.obs["_dataset"] == "ref",
            self.adata.obs["_labels_annotation"] != self.unknown_celltype_label,
        )
        if self.prediction_mode == "retrain" or "prediction_keys" not in self.adata.uns:
            self.adata.obs["_predict_cells"] = "relabel"
        else:
            self.adata.obs["_predict_cells"] = "reference"
            self.adata.obs.loc[self.adata.obs["_dataset"] == "query", "_predict_cells"] = "relabel"

        batch_count = self.adata.obs["_batch_annotation"].value_counts()
        if batch_count.min() < 11:
            logging.warning(
                f"Batch size of {batch_count.min()} is small. This will lead to issues when using BBKNN. Removing small batches."
            )
            valid_batches = batch_count[batch_count >= 11].index
            self.adata = self.adata[self.adata.obs["_batch_annotation"].isin(valid_batches)].copy()

        self.adata.obs["_labels_annotation"] = self.adata.obs["_labels_annotation"].astype("category")
        # Store values as default for current popv in adata
        self.adata.uns["unknown_celltype_label"] = self.unknown_celltype_label
        if self.prediction_mode == "retrain":
            self.label_categories = list(self.adata.obs["_labels_annotation"].cat.categories)
        self.adata.uns["label_categories"] = np.array(self.label_categories)
        self.adata.uns["_pretrained_scvi_path"] = self.pretrained_scvi_path
        self.adata.uns["_save_path_trained_models"] = self.save_path_trained_models
        self.adata.uns["_prediction_mode"] = self.prediction_mode
        self.adata.uns["_setup_dict"] = self.setup_dict
        self.adata.uns["_cl_obo_file"] = self.cl_obo_file
        self.adata.uns["_cl_ontology_file"] = self.cl_ontology_file
        self.adata.uns["_nlp_emb_file"] = self.nlp_emb_file
        if "prediction_keys" in self.adata.uns:
            self.adata.uns["ref_prediction_keys"] = self.adata.uns["prediction_keys"]
        else:
            self.adata.uns["ref_prediction_keys"] = []
        self.adata.uns["prediction_keys"] = []

        # Store some settings for reference models to output directory when retraining models.
        if self.prediction_mode == "retrain":
            data = {
                "gene_names": self.genes,
                "label_categories": self.label_categories,
            }

            with open(os.path.join(self.save_path_trained_models, "preprocessing.json"), "w") as f:
                json.dump(data, f, indent=4)



import itertools
from collections import Counter, defaultdict

import matplotlib.cm
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Polygon

# import bidi.algorithm  # for RTL languages


def plot(input_data, *args, **kwargs):
    at = AlluvialTool(input_data, *args, **kwargs)
    ax = at.plot(**kwargs)
    ax.axis("off")
    return ax


class AlluvialTool:
    def __init__(
        self,
        input_data=(),
        x_range=(0, 1),
        res=20,
        h_gap_frac=0.03,
        v_gap_frac=0.03,
        **kwargs,
    ):
        self.input = input_data
        self.x_range = x_range
        self.res = res  # defines the resolution of the splines for all veins
        self.combs = sorted(itertools.product((0, 1), (1, 0)), key=lambda xy: all(xy))
        self.trace_xy = self.make_vein_blueprint_xy_arrays()
        self.data_dic = self.read_input()
        self.item_widths_dic = self.get_item_widths_dic()
        self.a_members, self.b_members = self.get_item_groups(**kwargs)
        self.h_gap_frac = h_gap_frac
        self.h_gap = x_range[1] * h_gap_frac
        self.v_gap_frac = v_gap_frac
        self.v_gap = (
            sum([width for b_item_counter in self.data_dic.values() for width in b_item_counter.values()]) * v_gap_frac
        )
        self.group_widths = self.get_group_widths()
        self.item_coord_dic = self.make_item_coordinate_dic()
        self.alluvial_fan = self.generate_alluvial_fan()
        self.item_text_len, self.width_text_len = self.get_vein_label_lengths()

    def make_vein_blueprint_xy_arrays(self):
        y = np.array([0, 0.15, 0.5, 0.85, 1])
        x = np.linspace(self.x_range[0], self.x_range[-1], len(y))
        z = np.polyfit(x, y, 4)
        f = np.poly1d(z)

        blueprint_x_vals = np.linspace(x[0], x[-1], self.res)
        blueprint_y_vals = f(blueprint_x_vals)
        return blueprint_x_vals, blueprint_y_vals

    def get_vein_polygon_xy(self, y_range, width):
        x, y = self.trace_xy
        y0, yn = y_range
        scale = yn - y0
        ty = y * scale + y0
        x_new = np.concatenate(
            [
                x,
                x[::-1],
            ]
        )
        y_new = np.concatenate(
            [
                ty,
                ty[::-1] + width,
            ]
        )
        return np.array([x_new, y_new]).transpose()

    def read_input_from_list(self):
        data_table = np.array(self.input)
        assert len(set(data_table[:, 0]).intersection(set(data_table[:, 1]))) == 0, (
            "Column a and b have intersecting elements, this is not supported."
        )
        data_dic = defaultdict(Counter)
        for line in data_table:
            data_dic[line[0]][line[1]] += 1
        return data_dic

    def read_input_from_dict(self):
        # data_dic = self.input
        # data_table = []
        # for x_item, y_item_counter in data_dic.items():
        #     for y_item, count in y_item_counter.items():
        #         data_table += [[x_item, y_item]] * count
        # data_table = np.array(sorted(data_table))
        # return data_table, data_dic
        return self.input

    def read_input(self):
        if isinstance(self.input, dict):
            return self.read_input_from_dict()
        else:
            return self.read_input_from_list()

    def get_item_widths_dic(self):
        iwd = Counter()
        for a_item, b_item_counter in self.data_dic.items():
            for b_item, width in b_item_counter.items():
                iwd[a_item] += width
                iwd[b_item] += width
        return iwd

    def get_item_groups(self, a_sort=None, b_sort=None, **kwargs):
        _ = kwargs
        b_members = (
            sorted(
                {b_item for b_item_counter in self.data_dic.values() for b_item in b_item_counter},
                key=lambda x: self.item_widths_dic[x],
            )
            if not b_sort
            else b_sort
        )
        a_members = (
            sorted(
                set(self.data_dic.keys()),
                key=lambda x: self.item_widths_dic[x],
            )
            if not a_sort
            else a_sort
        )
        return a_members, b_members

    def get_group_widths(self):
        return [self.get_group_width(group) for group in (self.a_members, self.b_members)]

    def make_item_coordinate_dic(
        self,
    ):
        item_coord_dic = defaultdict(ItemCoordRecord)
        groups = self.a_members, self.b_members
        group_widths = self.group_widths
        for ind, group in enumerate(groups):
            last_pos = (max(group_widths) - group_widths[ind]) / 2
            for item in group:
                width = self.item_widths_dic[item]
                xy = (self.x_range[ind], last_pos)
                item_coord_dic[item].set_start_state(width, xy, side=ind)
                last_pos += width + self.v_gap
        return item_coord_dic

    def get_group_width(self, group):
        return sum([self.item_widths_dic[item] for item in group]) + (len(group) - 1) * self.v_gap

    def generate_alluvial_vein(self, a_item, b_item):
        width = self.data_dic[a_item][b_item]
        a_item_coord = self.item_coord_dic[a_item].read_state_and_advance_y(width)
        b_item_coord = self.item_coord_dic[b_item].read_state_and_advance_y(width)
        y_range = (
            a_item_coord[1],
            b_item_coord[1],
        )
        return self.get_vein_polygon_xy(y_range, width)

    def get_label_rectangles_xy(self, a_item, b_item):
        width = self.data_dic[a_item][b_item]
        return (
            self.generate_item_sub_rectangle(a_item, width),
            self.generate_item_sub_rectangle(b_item, width),
        )

    def generate_item_sub_rectangle(self, item, width):
        dic_entry = self.item_coord_dic[item]
        item_coord = dic_entry.read_state()
        sign = dic_entry.get_side_sign()
        return self.get_rectangle_xy(item_coord, width, sign)

    def get_rectangle_xy(self, item_coord, width, sign):
        x, y = item_coord
        rect = [
            [
                x + sign * 0.5 * (0.5 + xa) * self.h_gap,
                y + ya * width,
            ]
            for xa, ya in self.combs
        ]
        return np.array(rect)

    def generate_alluvial_fan(
        self,
    ):
        alluvial_fan = []
        for a_item in self.a_members:
            b_items4a_item = self.data_dic[a_item].keys()
            for b_item in self.b_members:
                if b_item in b_items4a_item:
                    l_a_rect, l_b_rect = self.get_label_rectangles_xy(a_item, b_item)
                    alluvial_fan += [
                        [
                            self.generate_alluvial_vein(a_item, b_item),
                            l_a_rect,
                            l_b_rect,
                            a_item,
                            b_item,
                        ]
                    ]
        return np.array(alluvial_fan, dtype=object)

    def plot(self, figsize=(10, 15), alpha=0.5, **kwargs):
        colors = self.get_color_array(**kwargs)
        fig, ax = plt.subplots(figsize=figsize)
        for num in (0, 1, 2):
            patches = [
                Polygon(
                    item,
                    facecolor=colors[ind],
                    alpha=alpha,
                )
                for ind, item in enumerate(self.alluvial_fan[:, num])
            ]
            for patch in patches:
                ax.add_patch(patch)
        self.auto_label_veins(**kwargs)
        self.label_sides(**kwargs)
        ax.autoscale()
        return ax

    def get_color_array(self, colors=None, color_side=0, rand_seed=1, cmap=None, **kwargs):
        _ = kwargs
        color_items = self.b_members if color_side else self.a_members
        lci = len(color_items)
        if rand_seed is not None:
            np.random.seed(rand_seed)
        cmap = cmap if cmap is not None else matplotlib.cm.get_cmap("hsv", lci * 10**3)
        color_array = colors if colors is not None else [cmap(item) for ind, item in enumerate(np.random.rand(lci))]
        ind_dic = {item: ind for ind, item in enumerate(color_items)}
        polygon_colors = []
        for (
            _,
            _,
            _,
            a_item,
            b_item,
        ) in self.alluvial_fan:
            item = b_item if color_side else a_item
            polygon_colors += [color_array[ind_dic[item]]]
        return np.array(polygon_colors)

    def get_vein_label_lengths(self):
        item_text_len = max([len(it) for it in self.item_widths_dic])
        width_text_len = max([len(str(w)) for w in self.item_widths_dic.values()])
        return item_text_len, width_text_len

    def auto_label_veins(self, fontname="Arial", **kwargs):
        # shift = max([len(item) for item in self.item_coord_dic.keys()]) / 50
        for item, vein in self.item_coord_dic.items():
            y_width = vein.get_width()
            sign = vein.get_side_sign()
            side = int(sign + 1) // 2
            ha = "left" if side else "right"
            plt.text(
                vein.get_x() + 1.5 * sign * self.h_gap,
                vein.get_y() + y_width / 2,
                self.item_text(item, side, **kwargs),
                ha=ha,
                va="center",
                fontname=fontname,
            )

    def label_sides(
        self,
        labels=None,
        label_shift=0,
        disp_width=False,
        wdisp_sep=7 * " ",
        fontname="Arial",
        **kwargs,
    ):
        if labels is not None:
            _ = kwargs
            y = max(self.group_widths) / 2
            itl, wtl = self.item_text_len, self.width_text_len
            for side, sign in enumerate((-1, 1)):
                plt.text(
                    self.x_range[side]
                    + sign * (label_shift + itl + int(disp_width) * (len(wdisp_sep) + wtl)) * self.h_gap_frac,
                    y,
                    labels[side],
                    # bidi.algorithm.get_display(labels[side]),  # RTL languages
                    ha="center",
                    va="center",
                    fontname=fontname,
                    fontsize=13,
                    rotation=90 - 180 * side,
                )

    def item_text(self, item, side, disp_width=False, wdisp_sep=7 * " ", width_in=True, **kwargs):
        _ = kwargs
        f_item = item
        # f_item = bidi.algorithm.get_display(item)  # for RTL languages
        tal = "<" if f_item == item else ">"
        if not disp_width:
            ans = f"{item:{tal}}"
        else:
            width = self.item_coord_dic[item].get_width()
            if (side and width_in) or (not side and not width_in):
                lc, rc, wl, wr, tl, tr = (
                    ">",
                    tal,
                    self.width_text_len,
                    self.item_text_len,
                    width,
                    f_item,
                )
            else:
                lc, rc, wl, wr, tl, tr = (
                    tal,
                    ">",
                    self.item_text_len,
                    self.width_text_len,
                    f_item,
                    width,
                )
            pat = f"{lc:{wl}d}{wdisp_sep}{rc:{wr}d}"
            ans = pat.format(
                tl,
                tr,
            )
        return ans


class ItemCoordRecord:
    def __init__(
        self,
    ):
        self.width = 0
        self.xy = ()
        self.curr_xy = self.xy[:]
        self.side = -1

    def set_start_state(self, width, xy, side):
        self.width = width
        self.xy = xy
        self.curr_xy = list(self.xy[:])
        self.side = side

    def read_state_and_advance_y(self, width):
        out = self.curr_xy[:]
        self.curr_xy[1] += width
        return out

    def read_state_and_advance_x(self, width):
        out = self.curr_xy[:]
        self.curr_xy[0] += width
        return out

    def read_state(self):
        return self.curr_xy[:]

    def get_xy(
        self,
    ):
        return self.xy

    def get_x(
        self,
    ):
        return self.xy[0]

    def get_y(
        self,
    ):
        return self.xy[1]

    def get_width(
        self,
    ):
        return self.width

    def get_side_sign(
        self,
    ):
        return 1 if self.side else -1



from __future__ import annotations

import logging
from pathlib import Path

import scanpy as sc
import scvi
from rich.console import Console
from rich.logging import RichHandler

popv_logger = logging.getLogger("popv")


class Config:
    """
    Config manager for PopV.

    Examples
    --------
    To set the seed

    >>> popv.settings.seed = 1

    To set the verbosity

    >>> import logging
    >>> popv.settings.verbosity = logging.INFO

    To set the logging directory

    >>> popv.settings.logging_dir = "./popv_log/"

    To set the number of jobs to be used

    >>> popv.settings.n_jobs = 2

    To set the number of largest dense dataset to be used

    >>> popv.settings.shard_size = 200000

    To enable cuml for rapid GPU based methods

    >>> popv.settings.cuml = True

    To recompute embeddings instead of using NN lookup

    >>> popv.settings.recompute_embeddings = True

    To not return probabilities for all classifiers

    >>> popv.settings.return_probabilities = False

    To not compute UMAP embedding for integration methods

    >>> popv.settings.compute_umap_embedding = False

    """

    def __init__(
        self,
        verbosity: int = logging.WARNING,
        seed: int | None = None,
        logging_dir: str = "./popv_log/",
        n_jobs: int = 1,
        cuml: bool = False,
        accelerator: str = "auto",
        shard_size: int = 100000,
        recompute_embeddings: bool = False,
        return_probabilities: bool = True,
        compute_umap_embedding: bool = True,
    ):
        """Set up Config manager for PopV."""
        self.seed = seed
        self.logging_dir = logging_dir
        self.verbosity = verbosity
        self.n_jobs = n_jobs
        self.cuml = cuml
        self.accelerator = accelerator
        self.shard_size = shard_size
        self.recompute_embeddings = recompute_embeddings
        self.return_probabilities = return_probabilities
        self.compute_umap_embedding = compute_umap_embedding

    @property
    def logging_dir(self) -> Path:
        """Directory for training logs (default `'./popv_log/'`)."""
        return self._logging_dir

    @logging_dir.setter
    def logging_dir(self, logging_dir: str | Path):
        self._logging_dir = Path(logging_dir).resolve()

    @property
    def n_jobs(self) -> int:
        """Jobs used for multiprocessing."""
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, n_jobs: int):
        """Random seed for torch and numpy."""
        sc.settings.n_jobs = n_jobs
        self._n_jobs = n_jobs

    @property
    def cuml(self) -> int:
        """Use RAPIDS and cuml."""
        return self._cuml

    @cuml.setter
    def cuml(self, cuml: bool):
        """Use RAPIDS and cuml."""
        self._cuml = cuml

    @property
    def shard_size(self) -> int:
        """Maximum number of cells in dense arrays."""
        return self._shard_size

    @shard_size.setter
    def shard_size(self, shard_size: int):
        """Maximum number of cells in dense arrays."""
        self._shard_size = shard_size

    @property
    def seed(self) -> int:
        """Random seed for torch and numpy."""
        return self._seed

    @seed.setter
    def seed(self, seed: int):
        """Random seed for torch and numpy."""
        scvi.settings.seed = seed
        self._seed = seed

    @property
    def verbosity(self) -> int:
        """Verbosity level (default `logging.INFO`)."""
        return self._verbosity

    @verbosity.setter
    def verbosity(self, level: str | int):
        """
        Set logging configuration for popV based on chosen level of verbosity.

        Parameters
        ----------
        level
            Sets "popV" logging level to `level`
        force_terminal
            Rich logging option, set to False if piping to file output.
        """
        self._verbosity = level
        popv_logger.setLevel(level)
        if len(popv_logger.handlers) == 0:
            console = Console(force_terminal=True)
            if console.is_jupyter is True:
                console.is_jupyter = False
            ch = RichHandler(level=level, show_path=False, console=console, show_time=False)
            formatter = logging.Formatter("%(message)s")
            ch.setFormatter(formatter)
            popv_logger.addHandler(ch)
        else:
            popv_logger.setLevel(level)

    @property
    def accelerator(self) -> bool:
        """Accelerator for scvi-tools models."""
        return self._accelerator

    @accelerator.setter
    def accelerator(self, accelerator: str):
        self._accelerator = accelerator

    @property
    def recompute_embeddings(self) -> bool:
        """Recompute UMAPs and BBKNN and Harmony."""
        return self._recompute_embeddings

    @recompute_embeddings.setter
    def recompute_embeddings(self, recompute_embeddings: bool):
        self._recompute_embeddings = recompute_embeddings

    @property
    def compute_umap_embedding(self) -> bool:
        """Compute UMAP embedding for integration methods."""
        return self._compute_umap_embedding

    @compute_umap_embedding.setter
    def compute_umap_embedding(self, compute_umap_embedding: bool):
        self._compute_umap_embedding = compute_umap_embedding

    @property
    def return_probabilities(self) -> bool:
        """Return internal certainties for all classifiers."""
        return self._return_probabilities

    @return_probabilities.setter
    def return_probabilities(self, return_probabilities: bool):
        self._return_probabilities = return_probabilities


settings = Config()



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class Random_Forest(BaseAlgorithm):
    """
    Class to compute Random forest classifier.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
        Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information.
        Default is "_labels_annotation".
    layer_key
        Key in layers field of adata used for classification. By default uses 'X' (log1p10K).
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_rf_prediction".
    enable_cuml
        Enable cuml, which currently doesn't support weighting. Default to popv.settings.cuml.
    classifier_dict
        Dictionary to supply non-default values for RF classifier. Options at :class:`sklearn.ensemble.RandomForestClassifier`.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        layer_key: str | None = None,
        result_key: str | None = "popv_rf_prediction",
        enable_cuml: bool = settings.cuml,
        classifier_dict: str | None = {},
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
        )
        self.layer_key = layer_key
        self.classifier_dict = {
            "class_weight": "balanced_subsample",
            "max_features": 200,
            "n_jobs": settings.n_jobs,
        }
        if classifier_dict is not None:
            self.classifier_dict.update(classifier_dict)
        self.enable_cuml = enable_cuml

    def predict(self, adata):
        """
        Predict celltypes using Random Forest.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Computing random forest classifier. Storing prediction in adata.obs["{self.result_key}"]')

        test_x = adata[adata.obs["_predict_cells"] == "relabel"].layers[self.layer_key] if self.layer_key else adata.X

        if adata.uns["_prediction_mode"] == "retrain":
            # CUML RF doesn't support pickling, we always use sklearn RF
            train_idx = adata.obs["_ref_subsample"]
            train_x = adata[train_idx].layers[self.layer_key] if self.layer_key else adata[train_idx].X
            train_y = adata.obs.loc[train_idx, self.labels_key].cat.codes.to_numpy()
            rf = RandomForestClassifier(**self.classifier_dict)
            rf.fit(train_x, train_y)
            joblib.dump(
                rf,
                open(
                    os.path.join(adata.uns["_save_path_trained_models"], "rf_classifier.joblib"),
                    "wb",
                ),
            )
        else:
            rf = joblib.load(
                open(
                    os.path.join(adata.uns["_save_path_trained_models"], "rf_classifier.joblib"),
                    "rb",
                )
            )

        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", self.result_key] = adata.uns["label_categories"][
            rf.predict(test_x)
        ]
        if self.return_probabilities:
            if f"{self.result_key}_probabilities" not in adata.obs.columns:
                adata.obs[f"{self.result_key}_probabilities"] = pd.Series(dtype="float64")
            adata.obs.loc[
                adata.obs["_predict_cells"] == "relabel",
                f"{self.result_key}_probabilities",
            ] = np.max(rf.predict_proba(test_x), axis=1).astype(float)



"""PopV."""

import sys
import warnings

if sys.version_info[:2] != (3, 10):
    warnings.warn(
        "Pretrained models on huggingface are trained with Python 3.11. "
        "Detected Python {sys.version.split()[:2]} will not load these models.",
        UserWarning,
        stacklevel=3,
    )

# Set default logging handler to avoid logging with logging.lastResort logger.
import logging

import scanpy as sc

from ._settings import settings  # isort: skip

# Import order to avoid circular imports
from . import algorithms, annotation, hub, preprocessing, visualization
from ._utils import create_ontology_resources

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata
package_name = "popv"
__version__ = importlib_metadata.version(package_name)

settings.verbosity = logging.INFO

# Jax sets the root logger, this prevents double output.
popv_logger = logging.getLogger("popv")
popv_logger.propagate = False


__all__ = [
    "algorithms",
    "annotation",
    "create_ontology_resources",
    "hub",
    "preprocessing",
    "settings",
    "visualization",
]



from ._base_algorithm import BaseAlgorithm
from ._bbknn import KNN_BBKNN
from ._celltypist import CELLTYPIST
from ._harmony import KNN_HARMONY
from ._onclass import ONCLASS
from ._rf import Random_Forest
from ._scanorama import KNN_SCANORAMA
from ._scanvi import SCANVI_POPV
from ._scvi import KNN_SCVI
from ._svm import Support_Vector
from ._xgboost import XGboost

__all__ = [
    "BaseAlgorithm",
    "CELLTYPIST",
    "KNN_BBKNN",
    "KNN_HARMONY",
    "KNN_SCANORAMA",
    "KNN_SCVI",
    "ONCLASS",
    "Random_Forest",
    "SCANVI_POPV",
    "Support_Vector",
    "XGboost",
]



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import pandas as pd
import scanpy as sc
from harmony import harmonize
from pynndescent import PyNNDescentTransformer
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class KNN_HARMONY(BaseAlgorithm):
    """
    Class to compute KNN classifier after Harmony integration.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information. Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information. Default is "_labels_annotation".
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_knn_harmony_prediction".
    embedding_key
        Key in obsm in which UMAP embedding of integrated data is stored.
        Default is "X_pca_harmony_popv".
    umap_key
        Key in obsm in which UMAP embedding of integrated data is stored.
        Default is "X_umap_harmony_popv".
    method_kwargs
        Additional parameters for HARMONY. Options at harmony.integrate_scanpy
    classifier_dict
        Dictionary to supply non-default values for KNN classifier. n_neighbors and weights supported.
    embedding_kwargs
        Dictionary to supply non-default values for UMAP embedding. Options at sc.tl.umap
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        result_key: str | None = "popv_knn_harmony_prediction",
        embedding_key: str | None = "X_pca_harmony_popv",
        umap_key: str | None = "X_umap_harmony_popv",
        method_kwargs: dict | None = None,
        classifier_dict: dict | None = None,
        embedding_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            embedding_key=embedding_key,
            umap_key=umap_key,
        )

        if embedding_kwargs is None:
            embedding_kwargs = {}
        if classifier_dict is None:
            classifier_dict = {}
        if method_kwargs is None:
            method_kwargs = {}

        self.method_kwargs = {"dimred": 50}
        if method_kwargs is not None:
            self.method_kwargs.update(method_kwargs)

        self.classifier_dict = {"weights": "uniform", "n_neighbors": 15}
        if classifier_dict is not None:
            self.classifier_dict.update(classifier_dict)

        self.embedding_kwargs = {"min_dist": 0.1}
        self.embedding_kwargs.update(embedding_kwargs)
        self.recompute_classifier = True

    def compute_integration(self, adata):
        logging.info("Integrating data with harmony")
        if (
            adata.uns["_prediction_mode"] == "inference"
            and self.embedding_key in adata.obsm
            and not settings.recompute_embeddings
        ):
            self.recompute_classifier = False
            index = joblib.load(os.path.join(adata.uns["_save_path_trained_models"], "pynndescent_index.joblib"))
            query_features = adata.obsm["X_pca"][adata.obs["_dataset"] == "query", :]
            indices, _ = index.query(query_features.astype(np.float32), k=5)
            neighbor_values = adata.obsm[self.embedding_key][adata.obs["_dataset"] == "ref", :][indices].astype(
                np.float32
            )
            adata.obsm[self.embedding_key][adata.obs["_dataset"] == "query", :] = np.mean(neighbor_values, axis=1)
            adata.obsm[self.embedding_key] = adata.obsm[self.embedding_key].astype(np.float32)
        elif adata.uns["_prediction_mode"] != "fast":
            adata.obsm[self.embedding_key] = harmonize(
                adata.obsm["X_pca"],
                adata.obs,
                batch_key=self.batch_key,
                use_gpu=settings.accelerator == "gpu",
            )
        else:
            raise ValueError(f"Prediction mode {adata.uns['_prediction_mode']} not supported for HARMONY")

    def predict(self, adata):
        """
        Predict celltypes using KNN on Harmony.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving knn on harmony results to adata.obs["{self.result_key}"]')

        if self.recompute_classifier:
            ref_idx = adata.obs["_labelled_train_indices"]

            train_X = adata[ref_idx].obsm[self.embedding_key].copy()
            train_Y = adata.obs.loc[ref_idx, self.labels_key].cat.codes.to_numpy()
            knn = make_pipeline(
                PyNNDescentTransformer(
                    n_neighbors=self.classifier_dict["n_neighbors"],
                    n_jobs=settings.n_jobs,
                ),
                KNeighborsClassifier(metric="precomputed", weights=self.classifier_dict["weights"]),
            )
            knn.fit(train_X, train_Y)
            if adata.uns["_prediction_mode"] == "retrain" and adata.uns["_save_path_trained_models"]:
                joblib.dump(
                    knn,
                    open(
                        os.path.join(
                            adata.uns["_save_path_trained_models"],
                            "harmony_knn_classifier.joblib",
                        ),
                        "wb",
                    ),
                )
        else:
            knn = joblib.load(
                open(
                    os.path.join(
                        adata.uns["_save_path_trained_models"],
                        "harmony_knn_classifier.joblib",
                    ),
                    "rb",
                )
            )

        # save_results
        embedding = adata[adata.obs["_predict_cells"] == "relabel"].obsm[self.embedding_key]
        knn_pred = knn.predict(embedding)
        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", self.result_key] = adata.uns["label_categories"][
            knn_pred
        ]
        if self.return_probabilities:
            if f"{self.result_key}_probabilities" not in adata.obs.columns:
                adata.obs[f"{self.result_key}_probabilities"] = pd.Series(dtype="float64")
            adata.obs.loc[
                adata.obs["_predict_cells"] == "relabel",
                f"{self.result_key}_probabilities",
            ] = np.max(embedding, axis=1)

    def compute_umap(self, adata):
        """
        Compute UMAP embedding of integrated data.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obsm[self.umap_key].
        """
        if self.compute_umap_embedding:
            logging.info(f'Saving UMAP of harmony results to adata.obs["{self.umap_key}"]')
            transformer = "rapids" if settings.cuml else None
            sc.pp.neighbors(adata, use_rep=self.embedding_key, transformer=transformer)
            method = "rapids" if settings.cuml else "umap"
            adata.obsm[self.umap_key] = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs).obsm[
                "X_umap"
            ]



from __future__ import annotations

import logging
import os

import numpy as np
import pandas as pd
import xgboost as xgb

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class XGboost(BaseAlgorithm):
    """
    Class to compute XGboost classifier.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
    labels_key
        Key in obs field of adata for cell-type information.
    layer_key
        Key in layers field of adata used for classification. By default uses 'X' (log1p10K).
    result_key
        Key in obs in which celltype annotation results are stored.
    classifier_dict
        Dictionary to supply non-default values for XGboost classifier.
        Options at :func:`xgboost.train`.
        Default is {'tree_method': 'hist', 'device': 'cuda' if settings.cuml else 'cpu', 'objective': 'multi:softprob'}.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        layer_key: str | None = None,
        result_key: str | None = "popv_xgboost_prediction",
        classifier_dict: str | None = {},
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
        )

        self.layer_key = layer_key
        self.classifier_dict = {
            "tree_method": "hist",
            "device": "cuda" if settings.cuml else "cpu",
            "objective": "multi:softprob",
        }
        if classifier_dict is not None:
            self.classifier_dict.update(classifier_dict)

    def predict(self, adata):
        """
        Predict celltypes using XGboost.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Computing XGboost classifier. Storing prediction in adata.obs["{self.result_key}"]')

        subset = adata[adata.obs["_predict_cells"] == "relabel"]
        test_x = subset.layers[self.layer_key] if self.layer_key else subset.X
        test_y = subset.obs[self.labels_key].cat.codes.to_numpy()
        dtest = xgb.DMatrix(test_x, test_y)

        if adata.uns["_prediction_mode"] == "retrain":
            train_idx = adata.obs["_ref_subsample"]
            train_x = adata[train_idx].layers[self.layer_key] if self.layer_key else adata[train_idx].X
            train_y = adata.obs.loc[train_idx, self.labels_key].cat.codes.to_numpy()
            dtrain = xgb.DMatrix(train_x, train_y)
            self.classifier_dict["num_class"] = len(adata.uns["label_categories"])

            bst = xgb.train(self.classifier_dict, dtrain, num_boost_round=300)
            bst.save_model(os.path.join(adata.uns["_save_path_trained_models"], "xgboost_classifier.model"))
        else:
            bst = xgb.Booster({"device": "cuda" if False else "cpu"})
            bst.load_model(os.path.join(adata.uns["_save_path_trained_models"], "xgboost_classifier.model"))

        output_probabilities = bst.predict(dtest)
        unassigned_idx = list(adata.uns["label_categories"]).index(adata.uns["unknown_celltype_label"])
        output_probabilities[:, unassigned_idx] = 0.0
        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", self.result_key] = adata.uns["label_categories"][
            np.argmax(output_probabilities, axis=1)
        ]
        if self.return_probabilities:
            if f"{self.result_key}_probabilities" not in adata.obs.columns:
                adata.obs[f"{self.result_key}_probabilities"] = pd.Series(dtype="float64")
            adata.obs.loc[
                adata.obs["_predict_cells"] == "relabel",
                f"{self.result_key}_probabilities",
            ] = np.max(output_probabilities, axis=1).astype(float)



from __future__ import annotations

import logging
import os

import numpy as np
import pandas as pd
import scipy
from OnClass.OnClassModel import OnClassModel

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class ONCLASS(BaseAlgorithm):
    """
    Class to compute OnClass cell-type prediction.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
        Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information.
        Default is "_labels_annotation".
    layer_key
        Layer in adata used for Onclass prediction.
        Default is adata.X.
    max_iter
        Maximum iteration in Onclass training.
        Default is 30.
    cell_ontology_obs_key
        Key in obs in which ontology celltypes are stored.
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_onclass_prediction".
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        layer_key: str | None = None,
        max_iter: int | None = 30,
        cell_ontology_obs_key: str | None = None,
        result_key: str | None = "popv_onclass_prediction",
        seen_result_key: str | None = "popv_onclass_seen",
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            seen_result_key=seen_result_key,
        )
        self.layer_key = layer_key
        self.cell_ontology_obs_key = cell_ontology_obs_key
        self.max_iter = max_iter
        self.labels_key = labels_key

    def predict(self, adata):
        """
        Predict celltypes using OnClass.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Computing Onclass. Storing prediction in adata.obs["{self.result_key}"]')
        adata.obs.loc[adata.obs["_dataset"] == "query", "self.labels_key"] = adata.uns["unknown_celltype_label"]

        train_idx = adata.obs["_ref_subsample"]

        if self.layer_key is None:
            train_x = adata[train_idx].X.copy()
        else:
            train_x = adata[train_idx].layers[self.layer_key].copy()
        if scipy.sparse.issparse(train_x):
            train_x = train_x.todense()

        cl_ontology_file = adata.uns["_cl_ontology_file"]
        nlp_emb_file = adata.uns["_nlp_emb_file"]
        train_model = OnClassModel(cell_type_nlp_emb_file=nlp_emb_file, cell_type_network_file=cl_ontology_file)

        if adata.uns["_save_path_trained_models"] is not None:
            model_path = os.path.join(adata.uns["_save_path_trained_models"], "OnClass")
        else:
            model_path = None

        if adata.uns["_prediction_mode"] == "retrain":
            train_y = adata[train_idx].obs[self.labels_key]
            _ = train_model.EmbedCellTypes(train_y)

            (
                corr_train_feature,
                corr_train_genes,
            ) = train_model.ProcessTrainFeature(
                train_x,
                train_y,
                adata.var_names,
                log_transform=False,
            )

            train_model.BuildModel(ngene=len(corr_train_genes))
            train_model.Train(
                corr_train_feature,
                train_y,
                save_model=model_path,
                max_iter=self.max_iter,
            )
        else:
            train_model.BuildModel(ngene=None, use_pretrain=model_path)

        subset = adata[adata.obs["_predict_cells"] == "relabel"]
        if self.layer_key is None:
            test_x = subset.X.copy()
        else:
            test_x = subset.layers[self.layer_key].copy()
        if self.return_probabilities:
            required_columns = {
                self.seen_result_key: pd.Series(index=subset.obs_names, dtype=str),
                self.result_key: pd.Series(index=subset.obs_names, dtype=str),
                f"{self.result_key}_probabilities": pd.Series(index=subset.obs_names, dtype=float),
                f"{self.seen_result_key}_probabilities": pd.Series(index=subset.obs_names, dtype=float),
            }
        else:
            required_columns = {
                self.seen_result_key: pd.Series(index=subset.obs_names, dtype=str),
                self.result_key: pd.Series(index=subset.obs_names, dtype=str),
            }

        result_df = pd.DataFrame(required_columns)
        shard_size = int(settings.shard_size)
        for i in range(0, subset.n_obs, shard_size):
            tmp_x = test_x[i : i + shard_size]
            if scipy.sparse.issparse(test_x):
                tmp_x = tmp_x.todense()
            names_x = subset.obs_names[i : i + shard_size]
            corr_test_feature = train_model.ProcessTestFeature(
                test_feature=tmp_x,
                test_genes=subset.var_names,
                use_pretrain=model_path,
                log_transform=False,
            )

            if adata.uns["_prediction_mode"] == "fast":
                onclass_pred = train_model.Predict(
                    corr_test_feature,
                    use_normalize=False,
                    refine=False,
                    unseen_ratio=-0.0,
                )
                onclass_seen = np.argmax(onclass_pred, axis=1)
                pred_label_str = [train_model.i2co[ind] for ind in onclass_seen]
                result_df.loc[names_x, self.result_key] = pred_label_str
                result_df.loc[names_x, self.seen_result_key] = pred_label_str
                if self.return_probabilities:
                    result_df.loc[names_x, f"{self.result_key}_probabilities"] = np.max(onclass_pred, axis=1)
                    result_df.loc[names_x, f"{self.seen_result_key}_probabilities"] = np.max(onclass_pred, axis=1)
            else:
                onclass_pred = train_model.Predict(
                    corr_test_feature,
                    use_normalize=False,
                    refine=True,
                    unseen_ratio=-1.0,
                )
                pred_label_str = [train_model.i2co[ind] for ind in onclass_pred[2]]
                result_df.loc[names_x, self.result_key] = pred_label_str

                onclass_seen = np.argmax(onclass_pred[0], axis=1)
                pred_label_str = [train_model.i2co[ind] for ind in onclass_seen]
                result_df.loc[names_x, self.seen_result_key] = pred_label_str

                if self.return_probabilities:
                    result_df.loc[names_x, f"{self.result_key}_probabilities"] = np.max(
                        onclass_pred[1], axis=1
                    ) / onclass_pred[1].sum(1)
                    result_df.loc[names_x, f"{self.seen_result_key}_probabilities"] = np.max(onclass_pred[0], axis=1)
        for col in required_columns.keys():
            if col not in adata.obs.columns:
                if "probabilities" in col:
                    adata.obs[col] = pd.Series(dtype="float64")  # Set dtype to float
                else:
                    adata.obs[col] = adata.uns["unknown_celltype_label"]
                    adata.obs[col] = adata.obs[col].astype(str)  # Set dtype to string
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", result_df.columns] = result_df



from scvi.utils import error_on_missing_dependencies

error_on_missing_dependencies("huggingface_hub")


from ._metadata import HubModelCardHelper, HubMetadata  # noqa
from ._model import HubModel  # noqa
from ._get_metrics import create_criticism_report  # noqa

__all__ = [
    "HubMetadata",
    "HubModel",
    "HubModelCardHelper",
    "create_criticism_report",
]



from __future__ import annotations

import logging
import os

import celltypist
import joblib
import numpy as np
import pandas as pd
import scanpy as sc
from scipy.stats import mode

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class CELLTYPIST(BaseAlgorithm):
    """
    Class to compute Celltypist classifier.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
        Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information.
        Default is "_labels_annotation".
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_celltypist_prediction".
    method_kwargs
        Additional parameters for celltypist training.
        Options at :func:`celltypist.train`.
    classifier_dict
        Dictionary to supply non-default values for celltypist annotation.
        Options at :func:`celltypist.annotate`.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        result_key: str | None = "popv_celltypist_prediction",
        method_kwargs: dict | None = None,
        classifier_dict: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
        )

        if classifier_dict is None:
            classifier_dict = {}
        if method_kwargs is None:
            method_kwargs = {}

        self.method_kwargs = {"check_expression": False, "n_jobs": 10, "max_iter": 500}
        if method_kwargs is not None:
            self.method_kwargs.update(method_kwargs)

        self.classifier_dict = {"mode": "best match", "majority_voting": True}
        if classifier_dict is not None:
            self.classifier_dict.update(classifier_dict)

    def predict(self, adata):
        """
        Predict celltypes using Celltypist.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving celltypist results to adata.obs["{self.result_key}"]')

        if adata.uns["_prediction_mode"] == "fast":
            self.classifier_dict["majority_voting"] = False
            over_clustering = None
        elif (
            adata.uns["_prediction_mode"] == "inference"
            and "over_clustering" in adata.obs
            and not settings.recompute_embeddings
        ):
            index = joblib.load(os.path.join(adata.uns["_save_path_trained_models"], "pynndescent_index.joblib"))
            query_features = adata.obsm["X_pca"][adata.obs["_dataset"] == "query", :]
            indices, _ = index.query(query_features.astype(np.float32), k=5)
            neighbor_values = adata.obs.loc[adata.obs["_dataset"] == "ref", "over_clustering"].cat.codes.values[indices]
            adata.obs.loc[adata.obs["_dataset"] == "query", "over_clustering"] = adata.obs[
                "over_clustering"
            ].cat.categories[mode(neighbor_values, axis=1).mode.flatten()]
            over_clustering = adata.obs.loc[adata.obs["_predict_cells"] == "relabel", "over_clustering"]
        else:
            transformer = "rapids" if settings.cuml else None
            sc.pp.neighbors(adata, n_neighbors=15, use_rep="X_pca", transformer=transformer)
            sc.tl.leiden(adata, resolution=25.0, key_added="over_clustering")
            over_clustering = adata.obs.loc[adata.obs["_predict_cells"] == "relabel", "over_clustering"]

        if adata.uns["_prediction_mode"] == "retrain":
            train_idx = adata.obs["_ref_subsample"]
            if len(train_idx) > 100000 and not settings.cuml:
                self.method_kwargs["use_SGD"] = True
                self.method_kwargs["mini_batch"] = True

            train_adata = adata[train_idx].copy()
            model = celltypist.train(
                train_adata,
                self.labels_key,
                use_GPU=settings.cuml,
                **self.method_kwargs,
            )

            model.write(os.path.join(adata.uns["_save_path_trained_models"], "celltypist.pkl"))
        predictions = celltypist.annotate(
            adata[adata.obs["_predict_cells"] == "relabel"],
            model=os.path.join(adata.uns["_save_path_trained_models"], "celltypist.pkl"),
            over_clustering=over_clustering,
            **self.classifier_dict,
        )
        out_column = (
            "majority_voting" if "majority_voting" in predictions.predicted_labels.columns else "predicted_labels"
        )

        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", self.result_key] = predictions.predicted_labels[
            out_column
        ]
        if self.return_probabilities:
            if f"{self.result_key}_probabilities" not in adata.obs.columns:
                adata.obs[f"{self.result_key}_probabilities"] = pd.Series(dtype="float64")
            adata.obs.loc[
                adata.obs["_predict_cells"] == "relabel",
                f"{self.result_key}_probabilities",
            ] = predictions.probability_matrix.max(axis=1).values



from __future__ import annotations

import json
import logging
import os
from pathlib import Path

import networkx as nx
import numpy as np
import pandas as pd
from anndata import AnnData
from scipy.sparse import csr_matrix


def get_minified_adata(
    adata,
) -> AnnData:
    """Return a minified AnnData.

    Parameters
    ----------
    adata
        Original AnnData, of which we want to create a minified version.
    """
    adata = adata.copy()
    del adata.raw
    all_zeros = csr_matrix(adata.X.shape)
    X = all_zeros
    layers = {layer: all_zeros.copy() for layer in adata.layers}
    adata.X = X
    adata.layers = layers
    return adata


def create_ontology_nlp_emb(lbl2sent, output_path):
    """
    Create ontology embeddings using NLP and saves them to {output_path}/cl.ontology.nlp.emb.

    Parameters
    ----------
    lbl2sent
        Dictionary with label as key and description as value.
    output_path
        Path to save the embeddings.

    """
    from sentence_transformers import SentenceTransformer

    model = SentenceTransformer("all-mpnet-base-v2")

    sentences = list(lbl2sent.values())
    sentence_embeddings = model.encode(sentences)

    sent2vec = {}
    for label, embedding in zip(lbl2sent.keys(), sentence_embeddings, strict=True):
        sent2vec[label] = embedding

    output_file = os.path.join(output_path, "cl.ontology.nlp.emb")
    with open(output_file, "w") as fout:
        for label, vec in sent2vec.items():
            fout.write(label + "\t" + "\t".join(map(str, vec)) + "\n")


def create_ontology_resources(cl_obo_file):
    """
    Create ontology resources.

    Parameters
    ----------
    cl_obo_file

    Returns
    -------
    g
        Graph object
    id2name
        Dictionary of ontology id to celltype names
    name2id
        Dictionary of celltype names to ontology id
    """
    import json

    # Read the taxrank ontology
    with open(cl_obo_file) as f:
        graph = json.load(f)["graphs"][0]
    output_path = Path(cl_obo_file).parent
    popv_dict = {}
    popv_dict["nodes"] = [entry for entry in graph["nodes"] if entry["type"] == "CLASS" and entry.get("lbl", False)]
    popv_dict["lbl_sentence"] = {
        entry[
            "lbl"
        ]: f"{entry['lbl']}: {entry.get('meta', {}).get('definition', {}).get('val', '')} {' '.join(entry.get('meta', {}).get('comments', []))}"
        for entry in popv_dict["nodes"]
    }
    popv_dict["id_2_lbl"] = {entry["id"]: entry["lbl"] for entry in popv_dict["nodes"]}
    popv_dict["lbl_2_id"] = {entry["lbl"]: entry["id"] for entry in popv_dict["nodes"]}
    popv_dict["edges"] = [
        i
        for i in graph["edges"]
        if i["sub"].split("/")[-1][0:2] == "CL" and i["obj"].split("/")[-1][0:2] == "CL" and i["pred"] == "is_a"
    ]
    popv_dict["ct_edges"] = [
        [popv_dict["id_2_lbl"][i["sub"]], popv_dict["id_2_lbl"][i["obj"]]] for i in popv_dict["edges"]
    ]
    create_ontology_nlp_emb(popv_dict["lbl_sentence"], output_path)

    with open(f"{output_path}/cl_popv.json", "w") as f:
        json.dump(popv_dict, f, indent=4)
    children_edge_celltype_df = pd.DataFrame(popv_dict["ct_edges"])
    children_edge_celltype_df.to_csv(f"{output_path}/cl.ontology", sep="\t", header=False, index=False)


def subsample_dataset(
    adata,
    labels_key,
    n_samples_per_label=100,
    ignore_label=None,
):
    """
    Subsamples dataset per label to n_samples_per_label.

    If a label has fewer than n_samples_per_label examples, then will use
    all the examples. For labels in ignore_label, they won't be included
    in the resulting subsampled dataset.

    Parameters
    ----------
    adata
        AnnData object
    labels_key
        Key in adata.obs for label information
    n_samples_per_label
        Maximum number of samples to use per label
    ignore_label
        List of labels to ignore (not subsample).

    Returns
    -------
    Returns list of obs_names corresponding to subsampled dataset

    """
    sample_idx = []
    labels_counts = dict(adata.obs[labels_key].value_counts())

    logging.info(f"Sampling {n_samples_per_label} cells per label")

    for label in ignore_label:
        labels_counts.pop(label, None)

    for label in labels_counts.keys():
        label_locs = np.where(adata.obs[labels_key] == label)[0]
        if labels_counts[label] < n_samples_per_label:
            sample_idx.append(label_locs)
        else:
            label_subset = np.random.choice(label_locs, n_samples_per_label, replace=False)
            sample_idx.append(label_subset)
    sample_idx = np.concatenate(sample_idx)
    return adata.obs_names[sample_idx]


def check_genes_is_subset(ref_genes, query_genes):
    """
    Check whether query_genes is a subset of ref_genes.

    Parameters
    ----------
    ref_genes
        List of reference genes
    query_genes
        List of query genes

    Returns
    -------
    is_subset
        True if it is a subset, False otherwise.

    """
    if len(set(query_genes)) != len(query_genes):
        logging.warning("Genes in query_dataset are not unique.")

    if set(ref_genes).issubset(set(query_genes)):
        logging.info("All ref genes are in query dataset. Can use pretrained models.")
        is_subset = True
    else:
        logging.info("Not all reference genes are in query dataset. Set 'prediction_mode' to 'retrain'.")
        is_subset = False
    return is_subset


def make_batch_covariate(adata, batch_keys, new_batch_key):
    """
    Combine all the batches in batch_keys into a single batch. Save result into adata.obs['_batch'].

    Parameters
    ----------
    adata
        Anndata object
    batch_keys
        List of keys in adat.obs corresponding to batches
    """
    adata.obs[new_batch_key] = adata.obs[batch_keys].astype(str).sum(1).astype("category")


def calculate_depths(g):
    """
    Calculate depth of each node in a network.

    Parameters
    ----------
    g
        Graph object to compute path_length.

    Returns
    -------
    depths
        Dictionary containing depth for each node

    """
    depths = {}

    for node in g.nodes():
        path = nx.shortest_path_length(g, node)
        if "cell" not in path:
            logging.warning("Celltype not in DAG: ", node)
        else:
            depth = path["cell"]
        depths[node] = depth

    return depths


def make_ontology_dag(cl_obo_file, lowercase=False):
    """
    Construct a graph with all cell-types.

    Parameters
    ----------
    cl_obo_file
        File with all ontology cell-types.

    Returns
    -------
    g
        Graph containing all cell-types
    """
    with open(cl_obo_file) as f:
        cell_ontology = json.load(f)
    g = nx.DiGraph()
    g.add_edges_from(cell_ontology["ct_edges"])

    if not nx.is_directed_acyclic_graph(g):
        raise ValueError(f"Graph is not a Directed Acyclic Graph. {nx.find_cycle(g, orientation='original')}")

    if lowercase:
        mapping = {s: s.lower() for s in list(g.nodes)}
        g = nx.relabel_nodes(g, mapping)
    return g


def majority_vote(x):
    a, b = np.unique(x, return_counts=True)
    return a[np.argmax(b)]


def majority_count(x):
    _, b = np.unique(x, return_counts=True)
    return np.max(b)



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import scanpy as sc
from scipy.stats import mode
from sklearn.neighbors import KNeighborsClassifier

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class KNN_BBKNN(BaseAlgorithm):
    """
    Class to compute KNN classifier after BBKNN integration.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information. Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information. Default is "_labels_annotation".
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_knn_bbknn_prediction".
    umap_key
        Key in obsm in which UMAP embedding of integrated data is stored.
        Default is "X_umap_bbknn_popv".
    method_kwargs
        Additional parameters for BBKNN.
        See :func:`scanpy.external.pp.bbknn`.
        Default is {"metric": "euclidean", "approx": True, "n_pcs": 50, "neighbors_within_batch": 3, "use_annoy": False}.
    classifier_kwargs
        Dictionary to supply non-default values for KNN classifier.
        See :class:`sklearn.neighbors.KNeighborsClassifier`.
        Default is {"weights": "uniform", "n_neighbors": 15}.
    embedding_kwargs
        Dictionary to supply non-default values for UMAP embedding.
        See :func:`scanpy.tl.umap`.
        Default is {"min_dist": 0.1}.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        result_key: str | None = "popv_knn_bbknn_prediction",
        umap_key: str | None = "X_umap_bbknn_popv",
        method_kwargs: dict | None = None,
        classifier_kwargs: dict | None = None,
        embedding_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            umap_key=umap_key,
        )
        if embedding_kwargs is None:
            embedding_kwargs = {}
        if classifier_kwargs is None:
            classifier_kwargs = {}
        if method_kwargs is None:
            method_kwargs = {}

        self.method_kwargs = {
            "metric": "euclidean",
            "approx": True,
            "n_pcs": 50,
            "neighbors_within_batch": 3,
            "use_annoy": False,
        }
        self.method_kwargs.update(method_kwargs)

        self.classifier_kwargs = {"weights": "uniform", "n_neighbors": 15}
        if classifier_kwargs is not None:
            self.classifier_kwargs.update(classifier_kwargs)

        self.embedding_kwargs = {"min_dist": 0.1}
        self.embedding_kwargs.update(embedding_kwargs)

    def compute_integration(self, adata):
        """
        Compute BBKNN integration.

        Parameters
        ----------
        adata
            AnnData object. Modified inplace.
        """
        logging.info("Integrating data with bbknn")
        if (
            adata.uns["_prediction_mode"] == "inference"
            and "X_umap_bbknn" in adata.obsm
            and not settings.recompute_embeddings
        ):
            index = joblib.load(os.path.join(adata.uns["_save_path_trained_models"], "pynndescent_index.joblib"))
            query_features = adata.obsm["X_pca"][adata.obs["_dataset"] == "query", :]
            indices, _ = index.query(query_features.astype(np.float32), k=5)

            neighbor_embedding = adata.obsm["X_umap_bbknn"][adata.obs["_dataset"] == "ref", :][indices].astype(
                np.float32
            )
            adata.obsm[self.umap_key][adata.obs["_dataset"] == "query", :] = np.mean(neighbor_embedding, axis=1)
            adata.obsm[self.umap_key] = adata.obsm[self.umap_key].astype(np.float32)

            neighbor_probabilities = adata.obs[f"{self.result_key}_probabilities"][adata.obs["_dataset"] == "ref", :][
                indices
            ].astype(np.float32)
            adata.obs.loc[adata.obs["_dataset"] == "query", f"{self.result_key}_probabilities"] = np.mean(
                neighbor_probabilities, axis=1
            )

            neighbor_prediction = adata.obs[f"{self.result_key}"][adata.obs["_dataset"] == "ref", :][indices].astype(
                np.float32
            )
            adata.obs.loc[adata.obs["_dataset"] == "query", f"{self.result_key}"] = mode(neighbor_prediction, axis=1)
        else:
            if len(adata.obs[self.batch_key].unique()) > 100:
                logging.warning("Using PyNNDescent instead of FAISS as high number of batches leads to OOM.")
                self.method_kwargs["neighbors_within_batch"] = 1  # Reduce memory usage.
                self.method_kwargs["pynndescent_n_neighbors"] = 10  # Reduce memory usage.
                sc.external.pp.bbknn(
                    adata, batch_key=self.batch_key, use_faiss=False, use_rep="X_pca", **self.method_kwargs
                )
            else:
                sc.external.pp.bbknn(
                    adata, batch_key=self.batch_key, use_faiss=True, use_rep="X_pca", **self.method_kwargs
                )

    def predict(self, adata):
        """
        Predict celltypes using Celltypist.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving knn on bbknn results to adata.obs["{self.result_key}"]')

        distances = adata.obsp["distances"]
        ref_idx = adata.obs["_labelled_train_indices"]
        ref_dist_idx = np.where(ref_idx)[0]
        train_y = adata.obs.loc[ref_idx, self.labels_key].cat.codes.to_numpy()
        train_distances = distances[ref_dist_idx, :][:, ref_dist_idx]
        test_distances = distances[:, :][:, ref_dist_idx]

        # Make sure BBKNN found the required number of neighbors, otherwise reduce n_neighbors for KNN.
        smallest_neighbor_graph = np.min(
            [
                np.diff(test_distances.indptr).min(),
                np.diff(train_distances.indptr).min(),
            ]
        )
        if smallest_neighbor_graph < self.classifier_kwargs["n_neighbors"]:
            logging.warning(f"BBKNN found only {smallest_neighbor_graph} neighbors. Reduced neighbors in KNN.")
            self.classifier_kwargs["n_neighbors"] = smallest_neighbor_graph

        knn = KNeighborsClassifier(metric="precomputed", **self.classifier_kwargs)
        knn.fit(train_distances, y=train_y)
        adata.obs[self.result_key] = adata.uns["label_categories"][knn.predict(test_distances)]

        if self.return_probabilities:
            adata.obs[f"{self.result_key}_probabilities"] = np.max(knn.predict_proba(test_distances), axis=1)

    def compute_umap(self, adata):
        """
        Compute UMAP embedding of integrated data.

        Parameters
        ----------
        adata
            AnnData object. Results are stored in adata.obsm[self.umap_key].
        """
        if self.compute_umap_embedding:
            logging.info(f'Saving UMAP of bbknn results to adata.obs["{self.embedding_key}"]')
            if len(adata.obs[self.batch_key]) < 30 and settings.cuml:
                method = "rapids"
            else:
                logging.warning("Using UMAP instead of RAPIDS as high number of batches leads to OOM.")
                method = "umap"
                # RAPIDS not possible here as number of batches drastically increases GPU RAM.
            adata.obsm[self.umap_key] = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs).obsm[
                "X_umap"
            ]



from __future__ import annotations

import logging

import numpy as np
import scanpy as sc
from pynndescent import PyNNDescentTransformer
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class KNN_SCANORAMA(BaseAlgorithm):
    """
    Class to compute KNN classifier after Scanorama integration.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information. Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information. Default is "_labels_annotation".
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_knn_scanorama_prediction".
    embedding_key
        Key in obsm in which embedding of integrated data is stored.
        Default is "X_pca_scanorama_popv".
    umap_key
        Key in obsm in which UMAP embedding of integrated data is stored.
        Default is "X_umap_scanorama_popv".
    method_kwargs
        Additional parameters for SCANORAMA. Options at :func:`scanpy.external.pp.scanorama_integrate`.
    classifier_kwargs
        Dictionary to supply non-default values for KNN classifier.
        See :class:`sklearn.neighbors.KNeighborsClassifier`.
        Default is {"weights": "uniform", "n_neighbors": 15}.
    embedding_kwargs
        Dictionary to supply non-default values for UMAP embedding.
        See :func:`scanpy.tl.umap`.
        Default is {"min_dist": 0.1}.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        result_key: str | None = "popv_knn_scanorama_prediction",
        embedding_key: str | None = "X_pca_scanorama_popv",
        umap_key: str | None = "X_umap_scanorama_popv",
        method_kwargs: dict | None = None,
        classifier_kwargs: dict | None = None,
        embedding_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            embedding_key=embedding_key,
            umap_key=umap_key,
        )

        if embedding_kwargs is None:
            embedding_kwargs = {}
        if classifier_kwargs is None:
            classifier_kwargs = {}
        if method_kwargs is None:
            method_kwargs = {}

        self.method_kwargs = {}
        if method_kwargs is not None:
            self.method_kwargs.update(method_kwargs)

        self.classifier_kwargs = {"weights": "uniform", "n_neighbors": 15}
        if classifier_kwargs is not None:
            self.classifier_kwargs.update(classifier_kwargs)

        self.embedding_kwargs = {
            "min_dist": 0.1,
        }
        self.embedding_kwargs.update(embedding_kwargs)

    def compute_integration(self, adata):
        """
        Integrate data using SCANORAMA.

        Parameters
        ----------
        adata
            AnnData object. Results are stored in adata.obsm[self.embedding_key].
        """
        logging.info("Integrating data with scanorama")
        tmp = adata[adata.obs.sort_values(self.batch_key).index, :].copy()
        sc.external.pp.scanorama_integrate(
            tmp,
            key=self.batch_key,
            adjusted_basis=self.embedding_key,
            **self.method_kwargs,
        )
        adata.obsm[self.embedding_key] = tmp[adata.obs_names, :].obsm[self.embedding_key]

    def predict(self, adata):
        """
        Compute KNN classifier on SCANORAMA integrated data.

        Parameters
        ----------
        adata
            AnnData object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving knn on scanorama results to adata.obs["{self.result_key}"]')

        ref_idx = adata.obs["_labelled_train_indices"]
        train_X = adata[ref_idx].obsm[self.embedding_key]
        train_Y = adata.obs.loc[ref_idx, self.labels_key].cat.codes.to_numpy()

        knn = make_pipeline(
            PyNNDescentTransformer(n_neighbors=self.classifier_kwargs["n_neighbors"], n_jobs=settings.n_jobs),
            KNeighborsClassifier(metric="precomputed", weights=self.classifier_kwargs["weights"]),
        )

        knn.fit(train_X, train_Y)
        knn_pred = knn.predict(adata.obsm[self.embedding_key])

        # save_results
        adata.obs[self.result_key] = adata.uns["label_categories"][knn_pred]

        if self.return_probabilities:
            adata.obs[f"{self.result_key}_probabilities"] = np.max(
                knn.predict_proba(adata.obsm[self.embedding_key]), axis=1
            )

    def compute_umap(self, adata):
        """
        Compute UMAP embedding of integrated data.

        Parameters
        ----------
        adata
            AnnData object. Results are stored in adata.obsm[self.umap_key].
        """
        if self.compute_umap_embedding:
            logging.info(f'Saving UMAP of scanorama results to adata.obs["{self.embedding_key}"]')

            transformer = "rapids" if settings.cuml else None
            sc.pp.neighbors(adata, use_rep=self.embedding_key, transformer=transformer)
            method = "rapids" if settings.cuml else "umap"
            adata.obsm[self.umap_key] = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs).obsm[
                "X_umap"
            ]



import json
import os

import pandas as pd
from anndata import AnnData
from sklearn.metrics import f1_score


def _dataframe_to_markdown(df):
    # Create the header
    pretty_columns = [
        (
            "Consensus Prediction"
            if col == "popv_prediction"
            else (col[4:-11].replace("_", " ").capitalize() if col.startswith("popv") else col)
        )
        for col in df.columns
    ]
    header = "| Cell Type | " + " | ".join(pretty_columns) + " |"
    separator = "| --- | " + " | ".join("---" for _ in df.columns) + " |"

    # Format values and create rows, including the index
    rows = "\n".join(
        "| "
        + " | ".join([str(index)] + [f"{value:.2f}" if not isinstance(value, int) else f"{value}" for value in row])
        + " |"
        for index, row in zip(df.index, df.values, strict=True)
    )

    # Combine header, separator, and rows
    return f"{header}\n{separator}\n{rows}"


def create_criticism_report(
    adata: AnnData | None = None,
    label_key: str | None = "_reference_labels_annotation",
    save_folder: str | None = None,
) -> dict:
    """
    Compute and store accuracy metrics for a model.

    Parameters
    ----------
    adata
        AnnData to compute metrics on.
    label_key
        Key in adata.obs to use as cell type labels. If None, will use the original label key from
        the model.
    save_folder
        Path to folder for storing the metrics. Preferred to store in save_path folder of model.
    """
    predictions = adata.uns["prediction_keys"] + ["popv_prediction"]
    annotations = adata.obs[[*predictions, label_key]].astype("category")
    query_annotations = annotations.loc[adata.obs["_dataset"] == "query"]
    ref_annotations = annotations.loc[adata.obs["_dataset"] == "ref"]

    # Function to compute F1 score for each column
    def compute_f1_scores_per_label(df, ground_truth_col, labels, prediction_cols):
        results = pd.DataFrame(index=labels, columns=prediction_cols)
        # Compute F1 scores per label for each prediction column
        results = pd.DataFrame(index=labels, columns=["N cells", *prediction_cols])
        # Compute the cell count for each label
        label_counts = df[ground_truth_col].value_counts()
        results["N cells"] = [label_counts.get(label, 0) for label in labels]
        for col in prediction_cols:
            # Compare each label using vectorized logic
            for label in labels:
                y_true = df[ground_truth_col] == label
                y_pred = df[col] == label
                results.loc[label, col] = f1_score(y_true, y_pred)

        return results

    # Compute F1 scores
    labels = adata.obs[label_key].value_counts().index
    f1_scores_per_label_query = compute_f1_scores_per_label(query_annotations, label_key, labels, predictions)
    f1_scores_per_label_ref = compute_f1_scores_per_label(ref_annotations, label_key, labels, predictions)
    md_query_accuracy = _dataframe_to_markdown(f1_scores_per_label_query)
    md_ref_accuracy = _dataframe_to_markdown(f1_scores_per_label_ref)

    markdown_dict = {
        "query_accuracy": md_query_accuracy,
        "ref_accuracy": md_ref_accuracy,
    }

    save_path = os.path.join(save_folder, "accuracies.json")

    with open(save_path, "w") as f:
        json.dump(markdown_dict, f, indent=4)



"""Helper function to execute cell-type annotation and accumulate results."""

from __future__ import annotations

import inspect
import logging
import os
import string
from collections import defaultdict
from dataclasses import dataclass, field

import anndata
import joblib
import networkx as nx
import numpy as np
import pandas as pd
from tqdm import tqdm

from popv import _utils, algorithms


@dataclass
class AlgorithmsNT:
    """Dataclass to store all available algorithms."""

    OUTDATED_ALGORITHMS: tuple[str, ...] = (
        "Random_Forest",
        "KNN_SCANORAMA",
    )
    FAST_ALGORITHMS: tuple[str, ...] = (
        "KNN_SCVI",
        "SCANVI_POPV",
        "Support_Vector",
        "XGboost",
        "ONCLASS",
        "CELLTYPIST",
    )
    CURRENT_ALGORITHMS: tuple[str, ...] = field(init=False)
    ALL_ALGORITHMS: tuple[str, ...] = field(init=False)

    def __post_init__(self):
        self.CURRENT_ALGORITHMS = tuple(
            i[0]
            for i in inspect.getmembers(algorithms, inspect.isclass)
            if i[0] not in self.OUTDATED_ALGORITHMS and i[0] != "BaseAlgorithm"
        )
        self.ALL_ALGORITHMS = tuple(
            i[0] for i in inspect.getmembers(algorithms, inspect.isclass) if i[0] != "BaseAlgorithm"
        )


algorithms_nt = AlgorithmsNT()


def annotate_data(
    adata: anndata.AnnData,
    methods: list | None = None,
    save_path: str | None = None,
    methods_kwargs: dict | None = None,
) -> None:
    """
    Annotate an AnnData dataset preprocessed by :class:`popv.preprocessing.Process_Query` by using the annotation pipeline.

    Parameters
    ----------
    adata
        AnnData of query and reference cells. AnnData object after running :class:`popv.preprocessing.Process_Query`.
    methods_
        List of methods used for cell-type annotation. Defaults to all algorithms.
    save_path
        Path were annotated query data is saved. Defaults to None and is not saving data.
    methods_kwargs
        Dictionary, where keys are used methods and values contain non-default parameters.
        Default to empty-dictionary.
    """
    if save_path is not None and not os.path.exists(save_path):
        os.makedirs(save_path, exist_ok=True)
    methods = (
        methods
        if isinstance(methods, list)
        else (
            algorithms_nt.ALL_ALGORITHMS
            if methods == "all"
            else (
                algorithms_nt.FAST_ALGORITHMS
                if adata.uns["_prediction_mode"] == "fast"
                else algorithms_nt.CURRENT_ALGORITHMS
            )
        )
    )
    if adata.uns["ref_prediction_keys"] is not None and adata.uns["_prediction_mode"] == "inference":
        if not set(methods).issubset(adata.uns["ref_prediction_keys"]):
            missing_methods = set(methods) - set(adata.uns["ref_prediction_keys"])
            ValueError(
                f"Method {missing_methods} are not present in the reference data."
                "Use relabel_reference_cells=True in Process_Query or remove these methods."
            )
        adata.obs[adata.uns["ref_prediction_keys"]] = adata.obs[adata.uns["ref_prediction_keys"]].astype("object")

    if adata.uns["_cl_obo_file"] is False and "ONCLASS" in methods:
        methods = tuple(method for method in methods if method != "ONCLASS")

    methods_kwargs = methods_kwargs if methods_kwargs else {}

    all_prediction_keys = []
    all_prediction_keys_seen = []

    for method in tqdm(methods):
        current_method = getattr(algorithms, method)(**methods_kwargs.pop(method, {}))
        current_method.compute_integration(adata)
        current_method.predict(adata)
        current_method.compute_umap(adata)
        all_prediction_keys += [current_method.result_key]
        all_prediction_keys_seen += [current_method.seen_result_key]

    # Here we compute the consensus statistics
    logging.info(f"Using predictions {all_prediction_keys} for PopV consensus")
    adata.uns["prediction_keys"] = all_prediction_keys
    adata.uns["prediction_keys_seen"] = all_prediction_keys_seen
    adata.uns["methods"] = list(methods)
    adata.uns["method_kwargs"] = methods_kwargs
    compute_consensus(adata, all_prediction_keys_seen)
    # No ontology prediction if ontology is set to False.
    if adata.uns["_cl_obo_file"] is False:
        adata.obs[["popv_prediction", "popv_prediction_score"]] = adata.obs[
            ["popv_majority_vote_prediction", "popv_majority_vote_score"]
        ]
        adata.obs[["popv_parent"]] = adata.obs[["popv_majority_vote_prediction"]]
    else:
        ontology_vote_onclass(adata, all_prediction_keys)
        ontology_parent_onclass(adata, all_prediction_keys)

    if save_path is not None:
        prediction_save_path = os.path.join(save_path, "predictions.csv")
        adata[adata.obs._dataset == "query"].obs[
            [
                *all_prediction_keys,
                "popv_prediction",
                "popv_prediction_score",
                "popv_majority_vote_prediction",
                "popv_majority_vote_score",
                "popv_parent",
            ]
        ].to_csv(prediction_save_path)

        logging.info(f"Predictions saved to {prediction_save_path}")


def compute_consensus(adata: anndata.AnnData, prediction_keys: list) -> None:
    """
    Compute consensus prediction and statistics between all methods.

    Parameters
    ----------
    adata
        AnnData object
    prediction_keys
        Keys in adata.obs containing predicted cell_types.

    Returns
    -------
    Saves the consensus prediction in adata.obs['popv_majority_vote_prediction']
    Saves the consensus percentage between methods in adata.obs['popv_majority_vote_score']

    """
    consensus_prediction = adata.obs[prediction_keys].apply(_utils.majority_vote, axis=1)
    adata.obs["popv_majority_vote_prediction"] = consensus_prediction

    agreement = adata.obs[prediction_keys].apply(_utils.majority_count, axis=1)
    adata.obs["popv_majority_vote_score"] = agreement.values
    adata.obs["popv_majority_vote_score"] = adata.obs["popv_majority_vote_score"].astype("category")


def ontology_vote_onclass(
    adata: anndata.AnnData,
    prediction_keys: list,
    save_key: str | None = "popv_prediction",
):
    """
    Compute prediction using ontology aggregation method.

    Parameters
    ----------
    adata
        AnnData object
    prediction_keys
        Keys in adata.obs containing predicted cell_types.
    save_key
        Name of the field in adata.obs to store the consensus prediction.

    Returns
    -------
    Saves the consensus prediction in adata.obs[save_key]
    Saves the consensus percentage between methods in adata.obs[save_key + '_score']
    Saves the overlap in original prediction in
    """
    if adata.uns["_prediction_mode"] == "retrain":
        G = _utils.make_ontology_dag(adata.uns["_cl_obo_file"])
        if adata.uns["_save_path_trained_models"] is not None:
            joblib.dump(
                G,
                open(
                    os.path.join(adata.uns["_save_path_trained_models"], "obo_dag.joblib"),
                    "wb",
                ),
            )
    else:
        G = joblib.load(
            open(
                os.path.join(adata.uns["_save_path_trained_models"], "obo_dag.joblib"),
                "rb",
            )
        )

    cell_type_root_to_node = {}
    aggregate_ontology_pred = [None] * adata.n_obs
    depth = {"cell": 0}
    scores = [None] * adata.n_obs
    depths = [None] * adata.n_obs
    onclass_depth = [None] * adata.n_obs
    depth["cell"] = 0

    for ind, cell in enumerate(adata.obs.index):
        score = defaultdict(lambda: 0)
        score["cell"] = 0
        for pred_key in prediction_keys:
            cell_type = adata.obs[pred_key][cell]
            if not pd.isna(cell_type):
                if cell_type in cell_type_root_to_node:
                    root_to_node = cell_type_root_to_node[cell_type]
                else:
                    root_to_node = nx.descendants(G, cell_type)
                    cell_type_root_to_node[cell_type] = root_to_node
                    depth[cell_type] = len(nx.shortest_path(G, cell_type, "cell"))
                    for ancestor_cell_type in root_to_node:
                        depth[ancestor_cell_type] = len(nx.shortest_path(G, ancestor_cell_type, "cell"))
                if pred_key == "popv_onclass_prediction":
                    onclass_depth[ind] = depth[cell_type]
                    for ancestor_cell_type in root_to_node:
                        score[ancestor_cell_type] += 1
                score[cell_type] += 1
        # Find cell-type most present across all classifiers.
        # If tie then deepest in network.
        # If tie then last in alphabet, just to make it consistent across multiple cells.
        celltype_consensus = max(
            score,
            key=lambda k: (
                score[k],
                depth[k],
                26 - string.ascii_lowercase.index(cell_type[0].lower()),
            ),
        )
        aggregate_ontology_pred[ind] = celltype_consensus
        scores[ind] = score[celltype_consensus]
        depths[ind] = depth[celltype_consensus]
    adata.obs[save_key] = aggregate_ontology_pred
    adata.obs[f"{save_key}_score"] = scores
    adata.obs[f"{save_key}_depth"] = depths
    adata.obs[f"{save_key}_onclass_relative_depth"] = np.array(onclass_depth) - adata.obs[f"{save_key}_depth"]
    # Change numeric values to categoricals.
    adata.obs[[f"{save_key}_score", f"{save_key}_depth", f"{save_key}_onclass_relative_depth"]] = adata.obs[
        [f"{save_key}_score", f"{save_key}_depth", f"{save_key}_onclass_relative_depth"]
    ].astype("category")
    return adata


def ontology_parent_onclass(
    adata: anndata.AnnData,
    prediction_keys: list,
    save_key: str = "popv_parent",
    allowed_errors: int = 2,
):
    """
    Compute common parent consensus prediction using ontology accumulation.

    Parameters
    ----------
    adata
        AnnData object
    prediction_keys
        Keys in adata.obs containing predicted cell_types.
    save_key
        Name of the field in adata.obs to store the consensus prediction. Default to 'popv_parent'.
    allowed_errors
        How many misclassifications are allowed to find common ontology ancestor. Defaults to 2.

    Returns
    -------
    Saves the consensus prediction in adata.obs[save_key]
    Saves the consensus percentage between methods in adata.obs[save_key + '_score']
    Saves the overlap in original prediction in
    """
    if adata.uns["_prediction_mode"] == "retrain":
        G = _utils.make_ontology_dag(adata.uns["_cl_obo_file"])
        if adata.uns["_save_path_trained_models"] is not None:
            joblib.dump(
                G,
                open(
                    os.path.join(adata.uns["_save_path_trained_models"], "obo_dag.joblib"),
                    "wb",
                ),
            )
    else:
        G = joblib.load(
            open(
                os.path.join(adata.uns["_save_path_trained_models"], "obo_dag.joblib"),
                "rb",
            )
        )

    cell_type_root_to_node = {}
    aggregate_ontology_pred = []
    depth = {"cell": 0}
    for cell in adata.obs.index:
        score = defaultdict(lambda: 0)
        score_popv = defaultdict(lambda: 0)
        score["cell"] = 0
        for pred_key in prediction_keys:
            cell_type = adata.obs[pred_key][cell]
            if not pd.isna(cell_type):
                if cell_type in cell_type_root_to_node:
                    root_to_node = cell_type_root_to_node[cell_type]
                else:
                    root_to_node = nx.descendants(G, cell_type)
                    cell_type_root_to_node[cell_type] = root_to_node
                    depth[cell_type] = len(nx.shortest_path(G, cell_type, "cell"))
                    for ancestor_cell_type in root_to_node:
                        depth[ancestor_cell_type] = len(nx.shortest_path(G, ancestor_cell_type, "cell"))
                for ancestor_cell_type in list(root_to_node) + [cell_type]:
                    score[ancestor_cell_type] += 1
                score_popv[cell_type] += 1
        score = {key: min(len(prediction_keys) - allowed_errors, value) for key, value in score.items()}

        # Find ancestor most present and deepest across all classifiers.
        # If tie, then highest in original classifier.
        # If tie then last in alphabet, just to make it consistent across multiple cells.
        celltype_consensus = max(
            score,
            key=lambda k: (
                score[k],
                depth[k],
                score_popv[k],
                26 - string.ascii_lowercase.index(cell_type[0].lower()),
            ),
        )
        aggregate_ontology_pred.append(celltype_consensus)
    adata.obs[save_key] = aggregate_ontology_pred
    return adata



"""Function to plot cell-type annotation."""

from __future__ import annotations

import logging
import os

import matplotlib.backends.backend_pdf
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.metrics import confusion_matrix

from .reproducibility import _alluvial


def _sample_report(adata, cell_type_key, score_key, pred_keys):
    adata.obs["counts"] = np.zeros(len(adata.obs))
    _counts_adata = adata.obs.groupby([cell_type_key, score_key]).count()[["counts"]].reset_index()
    counts_adata = _counts_adata.pivot(cell_type_key, score_key, "counts")
    counts_adata = counts_adata.dropna()
    np_counts = counts_adata.dropna().to_numpy()
    row_sums = np_counts.sum(axis=1)
    new_matrix = np_counts / row_sums[:, np.newaxis]
    ax = (
        pd.DataFrame(data=new_matrix, index=counts_adata.index, columns=counts_adata.columns)
        .sort_values(7)
        .plot(kind="bar", stacked=True, figsize=(20, 7))
    )
    plt.title("Agreement per celltype", fontsize=16)
    ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    plt.show()
    abundance_adata = adata.obs.groupby(cell_type_key).count()[["counts"]]
    abundance_adata = abundance_adata.dropna()
    ax = abundance_adata.plot(kind="bar", figsize=(20, 7))
    plt.title("Cell type abundance plot", fontsize=16)
    ax.bar_label(ax.containers[0])
    plt.show()
    for key in pred_keys:
        counts_adata = adata.obs.groupby([key, cell_type_key]).count().reset_index().pivot(key, cell_type_key, "counts")

        np_counts = counts_adata.dropna().to_numpy()
        row_sums = np_counts.sum(axis=0)
        new_matrix = np_counts / row_sums[np.newaxis, :]
        new_index = [counts_adata.index[r] + " " + str(np.sum(np_counts[r])) for r in range(new_matrix.shape[0])]
        new_columns = [counts_adata.columns[c] + " " + str(np.sum(np_counts[:, c])) for c in range(new_matrix.shape[1])]
        input_data = pd.DataFrame(data=new_matrix, index=new_index, columns=new_columns).to_dict()
        cmap = matplotlib.cm.get_cmap("jet")
        sorted_index = np.array(new_index)[sorted(range(new_matrix.shape[0]), key=lambda r: np.sum(np_counts[r]))]
        sorted_columns = np.array(new_columns)[
            sorted(range(new_matrix.shape[1]), key=lambda c: np.sum(np_counts[:, c]))
        ]
        ax = _alluvial.plot(
            input_data,
            alpha=0.4,
            color_side=1,
            figsize=(5, 10),
            wdisp_sep=" " * 2,
            cmap=cmap,
            fontname="Monospace",
            label_shift=2,
            b_sort=list(sorted_index),
            a_sort=list(sorted_columns),
            labels=("Method", "Consensus"),
        )

        ax.set_title(key, fontsize=14, fontname="Monospace")
        plt.show()


def agreement_score_bar_plot(
    adata,
    popv_prediction_key: str | None = "popv_prediction",
    consensus_score_key: str | None = "popv_prediction_score",
    save_folder: str | None = None,
):
    """
    Create bar-plot of prediction scores in query cells after running popv.

    Parameters
    ----------
    adata
        AnnData object.
    popv_prediction_score
        Key in adata.obs for prediction scores.
    save_folder
        Path to a folder for storing the plot. Defaults to None and plot is not stored.

    Returns
    -------
    Returns axis of corresponding plot.

    """
    ct = adata.obs[popv_prediction_key]

    celltypes = ct.unique()
    mean_agreement = [
        np.mean(
            adata[
                np.logical_and(
                    adata.obs["_dataset"] == "query",
                    ct == x,
                )
            ]
            .obs[consensus_score_key]
            .astype(float)
        )
        for x in celltypes
    ]
    mean_agreement = pd.DataFrame([mean_agreement], index=["agreement"], columns=celltypes).T
    mean_agreement.dropna(inplace=True)
    mean_agreement = mean_agreement.sort_values("agreement", ascending=True)
    ax = mean_agreement.plot.bar(y="agreement", figsize=(15, 2), legend=False)
    plt.ylabel("Mean Agreement")
    plt.xticks(rotation=290, ha="left")
    if save_folder is not None:
        figpath = os.path.join(save_folder, "percelltype_agreement_barplot.pdf")
        plt.savefig(figpath, bbox_inches="tight")
    return ax


def prediction_score_bar_plot(
    adata,
    popv_prediction_score: str | None = "popv_prediction_score",
    save_folder: str | None = None,
):
    """
    Create bar-plot of prediction scores in query cells after running popv.

    Parameters
    ----------
    adata
        AnnData object.
    popv_prediction_score
        Key in adata.obs for prediction scores.
    save_folder
        Path to a folder for storing the plot. Defaults to None and plot is not stored.

    Returns
    -------
    Returns axis object of corresponding plot.

    """
    ax = adata[adata.obs["_dataset"] == "query"].obs[popv_prediction_score].value_counts().sort_index().plot.bar()

    ax.set_xlabel("Score")
    ax.set_ylabel("Frequency")
    ax.set_title("PopV Prediction Score")
    if save_folder is not None:
        save_path = os.path.join(save_folder, "prediction_score_barplot.pdf")
        ax.get_figure().savefig(save_path, bbox_inches="tight", dpi=300)
    return ax


def celltype_ratio_bar_plot(
    adata,
    popv_prediction: str | None = "popv_prediction",
    save_folder: str | None = None,
    normalize: bool = True,
):
    """
    Create bar-plot of celltype rations in query as well as reference cells after running popv.

    Parameters
    ----------
    adata
        AnnData object.
    popv_prediction
        Key in adata.obs for predictions.
    save_folder
        Path to a folder for storing the plot. Defaults to None and plot is not stored.
    normalize
        Plot relative cell-type abundance. Set to False to plot absolute abundance.

    Returns
    -------
    Returns axis object of corresponding plot.

    """
    labels = adata.obs[popv_prediction].astype(str)
    is_query = adata.obs._dataset == "query"
    cell_types = np.unique(labels)
    prop = pd.DataFrame(index=cell_types, columns=["ref", "query"])
    for x in cell_types:
        prop.loc[x, "query"] = np.sum(labels[is_query] == x)
        prop.loc[x, "ref"] = np.sum(labels[~is_query] == x)
    if normalize:
        prop = prop.div(prop.sum(axis=0), axis=1)

    ax = prop.loc[cell_types].plot(kind="bar", figsize=(len(cell_types) * 0.5, 4), logy=(not normalize))
    ax.set_ylabel("Celltype")
    ax.set_ylabel("Celltype Abundance")
    if save_folder is not None:
        save_path = os.path.join(save_folder, "celltype_prop_barplot.pdf")
        ax.get_figure().savefig(save_path, bbox_inches="tight", dpi=300)
    return ax


def make_agreement_plots(
    adata,
    prediction_keys: list,
    popv_prediction_key: str | None = "popv_prediction",
    save_folder: str | None = None,
    show: bool = True,
):
    """
    Create plot of confusion matrix for different popv methods and consensus prediction.

    Parameters
    ----------
    adata
        AnnData object.
    prediction_keys
        List with key for methods for which confusion matrix is computed.
    popv_prediction_key
        Key in adata.obs for consensus prediction.
    save_folder
        Path to a folder for storing the plot. Defaults to None and plot is not stored.
    show
        If True, the plot will be shown in the console. If False, the plot will not be shown.

    Returns
    -------
    Returns axis of corresponding plot.

    """
    # clear all existing figures first
    # or else this will interfere with the pdf saving capabilities
    fig_nums = plt.get_fignums()
    for num in fig_nums:
        plt.close(num)

    for method in prediction_keys:
        logging.info(f"Making confusion matrix for {method}")
        _prediction_eval(
            adata.obs[method],
            adata.obs[popv_prediction_key],
            name=method,
            x_label=method,
            y_label=popv_prediction_key,
            res_dir=save_folder,
            show=show,
        )


def _prediction_eval(
    pred,
    labels,
    name,
    x_label="",
    y_label="",
    res_dir="./",
    show=True,
):
    """Generate confusion matrix."""
    types, _ = np.unique(np.concatenate([labels, pred]), return_inverse=True)
    mtx = confusion_matrix(labels, pred, normalize="true")
    df = pd.DataFrame(mtx, columns=types, index=types)
    df = df.loc[np.unique(labels), np.unique(pred)]
    df = df.rename_axis(x_label, axis="columns")
    df = df.rename_axis(y_label)
    df = df[df.idxmax(axis=1).unique()]
    plt.figure(figsize=(15, 12))
    sns.heatmap(df, linewidths=0.005, cmap="OrRd")
    plt.tight_layout()
    plt.title(name)
    if res_dir is not None:
        output_pdf_fn = os.path.join(res_dir, f"confusion_matrices_{name}.pdf")
        pdf = matplotlib.backends.backend_pdf.PdfPages(output_pdf_fn)
        for fig in range(1, plt.gcf().number + 1):
            pdf.savefig(fig)
        pdf.close()
    if show:
        plt.show()



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import pandas as pd
from sklearn import svm
from sklearn.calibration import CalibratedClassifierCV

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class Support_Vector(BaseAlgorithm):
    """
    Class to compute LinearSVC.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
        Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information.
        Default is "_labels_annotation".
    layer_key
        Key in layers field of adata used for classification. By default uses 'X' (log1p10K).
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_svm_prediction".
    classifier_dict
        Dictionary to supply non-default values for SVM classifier. Options at
        :class:`sklearn.svm.LinearSVC`.
        Default is {'C': 1, 'max_iter': 5000, 'class_weight': 'balanced'}.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        layer_key: str | None = None,
        result_key: str | None = "popv_svm_prediction",
        classifier_dict: str | None = {},
        train_both: bool = False,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
        )

        self.layer_key = layer_key
        self.classifier_dict = {
            "C": 1,
            "max_iter": 5000,
            "class_weight": "balanced",
        }
        if classifier_dict is not None:
            self.classifier_dict.update(classifier_dict)
        self.train_both = train_both

    def predict(self, adata):
        """
        Predict celltypes using LinearSVC.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Computing support vector machine. Storing prediction in adata.obs["{self.result_key}"]')
        test_x = adata.layers[self.layer_key] if self.layer_key else adata.X

        if adata.uns["_prediction_mode"] == "retrain":
            train_idx = adata.obs["_ref_subsample"]
            train_x = adata[train_idx].layers[self.layer_key] if self.layer_key else adata[train_idx].X
            train_x = np.array(train_x.todense())
            train_y = adata.obs.loc[train_idx, self.labels_key].cat.codes.to_numpy()
            if settings.cuml:
                from cuml.svm import LinearSVC
                from sklearn.multiclass import OneVsRestClassifier

                self.classifier_dict["probability"] = self.return_probabilities
                clf = OneVsRestClassifier(LinearSVC(**self.classifier_dict))
                clf.fit(train_x, train_y)
                joblib.dump(
                    clf,
                    open(
                        os.path.join(
                            adata.uns["_save_path_trained_models"],
                            "svm_classifier_cuml.joblib",
                        ),
                        "wb",
                    ),
                )
                self.classifier_dict.pop("probability")
            if not settings.cuml or self.train_both:
                clf = CalibratedClassifierCV(svm.LinearSVC(**self.classifier_dict))
                clf.fit(train_x, train_y)
                joblib.dump(
                    clf,
                    open(
                        os.path.join(
                            adata.uns["_save_path_trained_models"],
                            "svm_classifier.joblib",
                        ),
                        "wb",
                    ),
                )

        if self.return_probabilities:
            required_columns = [self.result_key, f"{self.result_key}_probabilities"]
        else:
            required_columns = [self.result_key]

        result_df = pd.DataFrame(index=adata.obs_names, columns=required_columns, dtype=float)
        result_df[self.result_key] = result_df[self.result_key].astype("object")
        if settings.cuml:
            clf = joblib.load(
                open(
                    os.path.join(
                        adata.uns["_save_path_trained_models"],
                        "svm_classifier_cuml.joblib",
                    ),
                    "rb",
                )
            )
            shard_size = int(settings.shard_size)
            for i in range(0, adata.n_obs, shard_size):
                tmp_x = test_x[i : i + shard_size]
                names_x = adata.obs_names[i : i + shard_size]
                tmp_x = np.array(tmp_x.todense())
                result_df.loc[names_x, self.result_key] = adata.uns["label_categories"][clf.predict(tmp_x).astype(int)]
                if self.return_probabilities:
                    result_df.loc[names_x, f"{self.result_key}_probabilities"] = np.max(
                        clf.predict_proba(tmp_x), axis=1
                    ).astype(float)
        else:
            clf = joblib.load(
                open(
                    os.path.join(adata.uns["_save_path_trained_models"], "svm_classifier.joblib"),
                    "rb",
                )
            )
            result_df[self.result_key] = adata.uns["label_categories"][clf.predict(test_x)]
            if self.return_probabilities:
                result_df[f"{self.result_key}_probabilities"] = np.max(clf.predict_proba(test_x), axis=1)
        for col in required_columns:
            if col not in adata.obs.columns:
                adata.obs[col] = (
                    pd.Series(dtype="float64") if "probabilities" in col else adata.uns["unknown_celltype_label"]
                )
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", result_df.columns] = result_df



from __future__ import annotations

import json
import logging
import os
import warnings
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING

import cellxgene_census
import rich
import scanpy as sc
from huggingface_hub import ModelCard, snapshot_download
from rich.markdown import Markdown
from scvi import settings
from scvi.utils import dependencies

from popv.annotation import AlgorithmsNT, annotate_data
from popv.hub._metadata import HubMetadata, HubModelCardHelper
from popv.preprocessing import Process_Query

if TYPE_CHECKING:
    from anndata import AnnData

logger = logging.getLogger(__name__)


class HubModel:
    """Wrapper for :class:`~scvi.model.base.BaseModelClass` backed by HuggingFace Hub.

    Parameters
    ----------
    repo_name
        ID of the HuggingFace repo where this model is uploaded
    local_dir
        Local directory where the data and pre-trained model reside.
    metadata
        Dict or a path to a file on disk where this metadata can be read from.
    model_card
        The model card for this pre-trained model. Model card is a markdown file that describes the
        pre-trained model/data and is displayed on HuggingFace. This can be either an instance of
        :class:`~huggingface_hub.ModelCard` or an instance of :class:`~popv.hub.HubModelCardHelper`
        that wraps the model card or a path to a file on disk where the model card can be read
        from.
    """

    def __init__(
        self,
        local_dir: str,
        metadata: dict | str | None = None,
        repo_name: str | None = None,
        model_card: HubModelCardHelper | ModelCard | str | None = None,
        ontology_dir: str | None = None,
    ):
        self._local_dir = local_dir
        self._ontology_dir = ontology_dir
        self._repo_name = repo_name

        self._model_path = f"{self._local_dir}"
        self._adata_path = f"{self._local_dir}/ref_adata.h5ad"
        self._minified_adata_path = f"{self._local_dir}/minified_ref_adata.h5ad"

        # lazy load - these are not loaded until accessed
        self._model = None
        self._adata = None
        self._minfied_adata = None

        # get the metadata from the parameters or from the disk
        metadata_path = f"{self._local_dir}/metadata.json"
        if isinstance(metadata, HubMetadata):
            self._metadata = metadata
        elif isinstance(metadata, str) or os.path.isfile(metadata_path):
            path = metadata if isinstance(metadata, str) else metadata_path
            content = Path(path).read_text()
            content_dict = json.loads(content)
            self._metadata = HubMetadata(**content_dict)
        else:
            raise ValueError("No metadata found")

        # get the model card from the parameters or from the disk
        model_card_path = f"{self._local_dir}/README.md"
        if isinstance(model_card, HubModelCardHelper):
            self._model_card = model_card.model_card
        elif isinstance(model_card, ModelCard):
            self._model_card = model_card
        elif isinstance(model_card, str) or os.path.isfile(model_card_path):
            path = model_card if isinstance(model_card, str) else model_card_path
            content = Path(path).read_text()
            self._model_card = ModelCard(content)
        else:
            raise ValueError("No model card found")

    def save(self, overwrite: bool = False) -> None:
        """Save the model card and metadata to the model directory.

        Parameters
        ----------
        overwrite
            Whether to overwrite existing files.
        """
        card_path = os.path.join(self._local_dir, "README.md")
        if os.path.isfile(card_path) and not overwrite:
            raise FileExistsError(f"Model card already exists at {card_path}. To overwrite, pass `overwrite=True`.")
        self.model_card.save(card_path)

        metadata_path = os.path.join(self._local_dir, "_required_metadata.json")
        with open(metadata_path, "w") as f:
            json.dump(self.metadata, f, indent=4)

    def annotate_data(
        self,
        query_adata: AnnData,
        query_batch_key: str | None = None,
        save_path: str = "tmp",
        prediction_mode: str = "fast",
        methods: list | None = None,
        gene_symbols: str | None = None,
    ) -> AnnData:
        """Annotate the query data with the trained model.

        Parameters
        ----------
        query_adata
            The query data to annotate.
        query_batch_key
            The batch key in the query data.
        save_path
            Path to save the query models.
        prediction_mode
            The prediction mode to use. Either "fast" or "inference".
            "fast" will only predict on the query data,
            while "inference" will integrate query and reference data.
        gene_symbols
            Gene symbols given as query_adata.var_names.

        Returns
        -------
        AnnData
            The annotated data.
        """
        ref_adata = self.adata if prediction_mode == "retrain" else self.minified_adata
        setup_dict = self.metadata.setup_dict
        if gene_symbols is not None:
            print("SSSSSS")
            query_adata = self.map_genes(adata=query_adata, gene_symbols=gene_symbols)
        print("LLLLLL", self.local_dir, os.listdir(self.local_dir))

        concatenate_adata = Process_Query(
            query_adata,
            ref_adata,
            query_batch_key=query_batch_key,
            ref_labels_key=setup_dict["ref_labels_key"],
            ref_batch_key=setup_dict["ref_batch_key"],
            unknown_celltype_label=setup_dict["unknown_celltype_label"],
            save_path_trained_models=self._local_dir,
            cl_obo_folder=self.ontology_dir,
            prediction_mode=prediction_mode,
            n_samples_per_label=100,
            hvg=None,
        ).adata
        methods_ = self.metadata.methods
        if prediction_mode == "fast":
            methods_ = [method for method in methods_ if method in AlgorithmsNT.FAST_ALGORITHMS]
        if methods is not None:
            if not set(methods).issubset(methods_):
                ValueError(f"Method {set(methods) - set(methods_)} is not supported. Consider retraining models.")
            methods_ = methods
        methods_kwargs = self.metadata.method_kwargs
        annotate_data(
            concatenate_adata,
            save_path=f"{save_path}/popv_output",
            methods=methods,
            methods_kwargs=methods_kwargs,
        )

        return concatenate_adata

    @dependencies("huggingface_hub")
    def push_to_huggingface_hub(
        self,
        repo_name: str,
        repo_token: str | None = None,
        repo_create: bool = False,
        repo_create_kwargs: dict | None = None,
        collection_slug: str | None = None,
        **kwargs,
    ):
        """Push this model to HuggingFace.

        If the dataset is too large to upload to HuggingFace, this will raise an
        exception prompting the user to upload the data elsewhere. Otherwise, the
        data, model card, and metadata are all uploaded to the given model repo.

        Parameters
        ----------
        repo_name
            ID of the HuggingFace repo where this model needs to be uploaded
        repo_token
            HuggingFace API token with write permissions if None uses token in HfFolder.get_token()
        repo_create
            Whether to create the repo
        repo_create_kwargs
            Keyword arguments passed into :meth:`huggingface_hub.HfApi.create_repo` if
            ``repo_create=True``.
        collection_slug
            The internal name in HuggingFace for a dataset collection.
        **kwargs
            Additional keyword arguments passed into :meth:`huggingface_hub.HfApi.upload_file`.
        """
        from huggingface_hub import HfApi, HfFolder, add_collection_item, create_repo

        self._repo_name = repo_name

        if repo_token is None:
            repo_token = HfFolder.get_token()
        elif os.path.isfile(repo_token):
            repo_token = Path(repo_token).read_text()
        if repo_create:
            repo_create_kwargs = repo_create_kwargs or {}
            create_repo(repo_name, token=repo_token, **repo_create_kwargs)
        api = HfApi()
        # upload the model card
        self.model_card.push_to_hub(repo_name, token=repo_token)
        # upload the model
        api.upload_folder(
            folder_path=self._local_dir,
            repo_id=repo_name,
            token=repo_token,
            ignore_patterns="*h5ad",  # Ignore all h5ad files.
            **kwargs,
        )
        # upload the metadata
        api.upload_file(
            path_or_fileobj=json.dumps(asdict(self.metadata), indent=4).encode(),
            path_in_repo="metadata.json",
            repo_id=repo_name,
            token=repo_token,
            **kwargs,
        )
        if os.path.isfile(f"{self._local_dir}/minified_ref_adata.h5ad"):
            api.upload_file(
                path_or_fileobj=f"{self._local_dir}/minified_ref_adata.h5ad",
                path_in_repo="minified_ref_adata.h5ad",
                repo_id=repo_name,
                token=repo_token,
                **kwargs,
            )
        collection_slug = collection_slug

        if collection_slug is not None:
            add_collection_item(
                collection_slug=collection_slug,
                item_id=repo_name,
                item_type="model",
                exists_ok=True,
            )

    @classmethod
    def pull_from_huggingface_hub(
        cls,
        repo_name: str,
        cache_dir: str | None = None,
        revision: str | None = None,
        **kwargs,
    ):
        """Download the given model repo from HuggingFace.

        The model, its card, data, metadata are downloaded to a cached location on disk
        selected by HuggingFace and an instance of this class is created with that info
        and returned.

        Parameters
        ----------
        repo_name
            ID of the HuggingFace repo where this model needs to be uploaded
        cache_dir
            The directory where the downloaded model artifacts will be cached
        revision
            The revision to pull from the repo. This can be a branch name, a tag, or a full-length
            commit hash. If None, the default (latest) revision is pulled.
        kwargs
            Additional keyword arguments to pass to :func:`huggingface_hub.snapshot_download`.
        """
        if revision is None:
            warnings.warn(
                "No revision was passed, so the default (latest) revision will be used.",
                UserWarning,
                stacklevel=settings.warnings_stacklevel,
            )

        snapshot_folder = snapshot_download(
            repo_id=repo_name,
            cache_dir=cache_dir,
            revision=revision,
            **kwargs,
        )
        ontology_snapshot = snapshot_download(
            repo_id="popV/ontology",
            repo_type="dataset",
            cache_dir=cache_dir,
        )
        model_card = ModelCard.load(repo_name)
        return cls(
            snapshot_folder,
            model_card=model_card,
            repo_name=repo_name,
            ontology_dir=ontology_snapshot,
        )

    def __repr__(self):
        def eval_obj(obj):
            return "No" if obj is None else "Yes"

        print(
            "HubModel with:\n"
            f"local_dir: {self._local_dir}\n"
            f"model loaded? {eval_obj(self._model)}\n"
            f"adata loaded? {eval_obj(self._adata)}\n"
            f"metadata:\n{self.metadata}\n"
            f"model_card:"
        )
        rich.print(Markdown(self.model_card.content.replace("\n", "\n\n")))
        return ""

    @property
    def local_dir(self) -> str:
        """The local directory where the data and pre-trained model reside."""
        return self._local_dir

    @property
    def ontology_dir(self) -> str:
        """The local directory where the models are downloaded."""
        return self._ontology_dir

    @property
    def repo_name(self) -> str:
        """The local directory where the data and pre-trained model reside."""
        return self._repo_name

    @property
    def metadata(self) -> dict:
        """The metadata for this model."""
        return self._metadata

    @property
    def model_card(self) -> ModelCard:
        """The model card for this model."""
        return self._model_card

    @property
    def adata(self) -> AnnData | None:
        """Returns the full training data for this model.

        If the data has not been loaded yet, this will call :func:`cellxgene_census.download_source_h5ad`.
        Otherwise, it will simply return the loaded data.
        """
        if self._adata is None:
            cellxgene_census.download_source_h5ad(
                dataset_id=self.metadata.cellxgene_url.rsplit("/", 2)[1].rsplit(".")[0],
                census_version="latest",
                to_path=self._adata_path,
            )
            self._adata = sc.read_h5ad(self._adata_path)
        return self._adata

    @property
    def minified_adata(self) -> AnnData | None:
        """Returns the minified data for this model.

        If the data has not been loaded yet, this will call :func:`scanpy.read_h5ad`.
        Otherwise, it will simply return the loaded data.
        """
        if self._minfied_adata is None:
            self._minfied_adata = sc.read_h5ad(self._minified_adata_path)
        return self._minfied_adata

    def map_genes(self, adata, gene_symbols) -> AnnData | None:
        """Map genes to CELLxGENE census gene IDs."""
        with cellxgene_census.open_soma() as census:
            var_df = cellxgene_census.get_var(
                census,
                organism="homo_sapiens",
            )
            feature_dict = dict(zip(var_df[gene_symbols], var_df["feature_id"], strict=True))
        adata.var["old_index"] = adata.var_names
        adata.var_names = adata.var_names.map(feature_dict)
        adata = adata[:, adata.var.index.notna()].copy()
        return adata



from __future__ import annotations

import json
import os
import sys
from dataclasses import asdict, dataclass, field

from anndata import AnnData
from huggingface_hub import ModelCard, ModelCardData

from ._template import (
    template,
)


@dataclass
class HubMetadata:
    """Encapsulates the required metadata for `popV` hub models.

    Parameters
    ----------
    popv_version
        The version of `popV` that the model was trained with.
    anndata_version
        The version of anndata used during model training.
    setup_dict
        The setup dictionary used to preprocess the data.
    prediction_keys
        The keys used to store the predictions in the AnnData object.
    methods_
        The methods used to predict celltypes
    method_kwargs
        The keyword arguments used in the methods.
    cellxgene_url
        Link to the data in the CELLxGENE portal (viewer).
    """

    popv_version: str
    anndata_version: str
    setup_dict: dict
    prediction_keys: list[str]
    method_kwargs: dict
    methods: list[str]
    method_kwargs: dict
    cellxgene_url: str | None = None

    @classmethod
    def from_anndata(
        cls,
        adata: AnnData,
        anndata_version: str,
        popv_version: str,
        **kwargs,
    ):
        """Create a `HubMetadata` object from an AnnData file.

        Parameters
        ----------
        adata
            The AnnData object used to train the model.
        anndata_version
            The version of anndata used during model training.
        popv_version
            The version of `popV` that the model was trained with.
        kwargs
            Additional keyword arguments to pass to the HubMetadata initializer.
        """
        setup_dict = adata.uns["_setup_dict"]
        prediction_keys = adata.uns["prediction_keys"]
        methods = adata.uns["methods"]
        method_kwargs = adata.uns["method_kwargs"]

        return cls(
            popv_version=popv_version,
            anndata_version=anndata_version,
            setup_dict=setup_dict,
            prediction_keys=prediction_keys,
            methods=methods,
            method_kwargs=method_kwargs,
            **kwargs,
        )

    def save(self, save_path: str, overwrite: bool = False) -> None:
        """Save the metadata to a JSON file.

        Parameters
        ----------
        save_path
            The path to which to save the metadata as a JSON file.
        overwrite
            Whether to overwrite the file if it already exists.
        """
        if os.path.isfile(save_path) and not overwrite:
            raise FileExistsError(f"File already exists at {save_path}. To overwrite, pass `overwrite=True`.")
        with open(save_path, "w") as f:
            json.dump(asdict(self), f, indent=4)


@dataclass
class HubModelCardHelper:
    """A helper for creating a `ModelCard` for `popV` hub models.

    Parameters
    ----------
    license_info
        The license information for the model.
    scvi_version
        The version of `scvi-tools` that the model was trained with.
    anndata_version
        The version of anndata used during model training.
    tissues
        The tissues of the training data.
    cellxgene_url
        Link to the data in the CELLxGENE portal.
    description
        A description of the model.
    references_
        A list of references for the model.
    metrics_report
        A dictionary containing the metrics report for the model.

    Attributes
    ----------
    model_card : ModelCard
        Stores the model card.

    Notes
    -----
    It is not required to use this class to create a `ModelCard`. But this helps you do so in a way
    that is consistent with other `popV` hub models. You can think of this as a
    template. The resulting
    huggingface :class:`~huggingface_hub.ModelCard` can be accessed via the
    :attr:`~popv.hub.HubModelCardHelper.model_card` property.
    """

    license_info: str
    anndata_version: str
    tissues: list[str] = field(default_factory=list)
    cellxgene_url: str | None = None
    description: str = "To be added..."
    references: str = "To be added..."
    metrics_report: str | None = None
    training_code_url: str = "Not provided by uploader."

    def __post_init__(self):
        self.model_card = self._to_model_card()

    @classmethod
    def from_dir(
        cls,
        local_dir: str,
        license_info: str,
        anndata_version: str,
        metrics_report: str | None = None,
        **kwargs,
    ):
        """Create a `HubModelCardHelper` object from a local directory.

        Parameters
        ----------
        local_dir
            The local directory containing the model files.
        license_info
            The license information for the model.
        anndata_version
            The version of anndata used during model training.
        metrics_report
            Path to the json with stored metrics report.
        data_is_minified
            Whether the training data uploaded with the model has been minified.
        kwargs
            Additional keyword arguments to pass to the HubModelCardHelper initializer.
        """
        if metrics_report is None:
            if os.path.isfile(f"{local_dir}/accuracies.json"):
                with open(f"{local_dir}/accuracies.json") as f:
                    metrics_report = json.load(f)
            else:
                metrics_report = None
        else:
            with open(metrics_report) as f:
                metrics_report = json.load(f)

        return cls(
            license_info,
            anndata_version,
            metrics_report=metrics_report,
            **kwargs,
        )

    def _to_model_card(self) -> ModelCard:
        # define tags
        tags = [
            "biology",
            "genomics",
            "single-cell",
            f"anndata_version:{self.anndata_version}",
            f"python_version:{'.'.join([str(i) for i in sys.version_info[:3]])}",
            "popV",
        ]
        for t in self.tissues:
            tags.append(f"tissue: {t}")

        # define the card data, which is the header
        card_data = ModelCardData(
            license=self.license_info,
            library_name="popV",
            tags=tags,
        )

        if self.metrics_report is not None:
            validation_accuracies = self.metrics_report.get("query_accuracy", "Not provided by uploader.")
            train_accuracies = self.metrics_report.get("ref_accuracy", "Not provided by uploader.")
        else:
            validation_accuracies = "Not provided by uploader."
            train_accuracies = "Not provided by uploader."

        # create the content from the template
        content = template.format(
            card_data=card_data.to_yaml(),
            description=self.description,
            cellxgene_url=self.cellxgene_url,
            references=self.references,
            validation_accuracies=validation_accuracies,
            train_accuracies=train_accuracies,
            training_code_url=self.training_code_url,
        )

        # finally create and return the actual card
        return ModelCard(content)



from __future__ import annotations

import logging
import os

import numpy as np
import pandas as pd
import scanpy as sc
import scvi

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class SCANVI_POPV(BaseAlgorithm):
    """
    Class to compute classifier in scANVI model.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
        Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information.
        Default is "_labels_annotation".
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_scanvi_prediction".
    embedding_key
        Key in obsm in which latent embedding is stored.
        Default is "X_scanvi_popv".
    umap_key
        Key in obsm in which UMAP embedding of integrated data is stored.
        Default is "X_umap_scanvi_popv".
    model_kwargs
        Dictionary to supply non-default values for SCVI model. Options at :class:`scvi.model.SCANVI`.
    classifier_kwargs
        Dictionary to supply non-default values for SCANVI classifier.
        Options at classifier_paramerers in :class:`scvi.model.SCANVI`.
    embedding_kwargs
        Dictionary to supply non-default values for UMAP embedding. Options at :func:`scanpy.tl.umap`.
    train_kwargs
        Dictionary to supply non-default values for training scvi. Options at :meth:`scvi.model.SCANVI.train`.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        save_folder: str | None = None,
        result_key: str | None = "popv_scanvi_prediction",
        embedding_key: str | None = "X_scanvi_popv",
        umap_key: str | None = "X_umap_scanvi_popv",
        model_kwargs: dict | None = None,
        classifier_kwargs: dict | None = None,
        embedding_kwargs: dict | None = None,
        train_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            embedding_key=embedding_key,
        )
        self.umap_key = umap_key
        self.save_folder = save_folder

        if embedding_kwargs is None:
            embedding_kwargs = {}
        if classifier_kwargs is None:
            classifier_kwargs = {}
        if model_kwargs is None:
            model_kwargs = {}
        if train_kwargs is None:
            train_kwargs = {}

        self.model_kwargs = {
            "dropout_rate": 0.05,
            "dispersion": "gene",
            "n_layers": 3,
            "n_latent": 20,
            "gene_likelihood": "nb",
            "use_batch_norm": "none",
            "use_layer_norm": "both",
            "encode_covariates": True,
        }
        if model_kwargs is not None:
            self.model_kwargs.update(model_kwargs)

        self.train_kwargs = {
            "max_epochs": 20,
            "batch_size": 512,
            "n_samples_per_label": 20,
            "accelerator": settings.accelerator,
            "plan_kwargs": {"n_epochs_kl_warmup": 20},
            "max_epochs_unsupervised": 20,
        }
        self.train_kwargs.update(train_kwargs)
        self.max_epochs_unsupervised = self.train_kwargs.pop("max_epochs_unsupervised")
        self.max_epochs = self.train_kwargs.get("max_epochs", None)

        self.classifier_kwargs = {"n_layers": 3, "dropout_rate": 0.1}
        if classifier_kwargs is not None:
            self.classifier_kwargs.update(classifier_kwargs)

        self.embedding_kwargs = {"min_dist": 0.3}
        self.embedding_kwargs.update(embedding_kwargs)

    def compute_integration(self, adata):
        """
        Compute scANVI model and integrate data.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obsm[self.embedding_key].
        """
        logging.info("Integrating data with scANVI")
        if adata.uns["_prediction_mode"] == "retrain":
            if adata.uns["_pretrained_scvi_path"]:
                scvi_model = scvi.model.SCVI.load(
                    os.path.join(adata.uns["_save_path_trained_models"], "scvi"),
                    adata=adata,
                )
            else:
                scvi.model.SCVI.setup_anndata(
                    adata,
                    batch_key=self.batch_key,
                    labels_key=self.labels_key,
                    layer="scvi_counts",
                )
                scvi_model = scvi.model.SCVI(adata, **self.model_kwargs)
                scvi_model.train(
                    max_epochs=self.max_epochs_unsupervised,
                    accelerator=settings.accelerator,
                    plan_kwargs={"n_epochs_kl_warmup": 20},
                )

            self.model = scvi.model.SCANVI.from_scvi_model(
                scvi_model,
                unlabeled_category=adata.uns["unknown_celltype_label"],
                classifier_parameters=self.classifier_kwargs,
            )
        else:
            query = adata[adata.obs["_predict_cells"] == "relabel"].copy()
            self.model = scvi.model.SCANVI.load_query_data(
                query,
                os.path.join(adata.uns["_save_path_trained_models"], "scanvi"),
                freeze_classifier=True,
            )

        if adata.uns["_prediction_mode"] == "fast":
            self.train_kwargs.update({"max_epochs": 1})
        self.model.train(**self.train_kwargs)
        if adata.uns["_prediction_mode"] == "retrain":
            self.model.save(
                os.path.join(adata.uns["_save_path_trained_models"], "scanvi"),
                save_anndata=False,
                overwrite=True,
            )
        latent_representation = self.model.get_latent_representation()
        relabel_indices = adata.obs["_predict_cells"] == "relabel"
        if self.embedding_key not in adata.obsm:
            # Initialize X_scanvi with the correct shape if it doesn't exist
            adata.obsm[self.embedding_key] = np.zeros((adata.n_obs, latent_representation.shape[1]))
        adata.obsm[self.embedding_key][relabel_indices, :] = latent_representation

    def predict(self, adata):
        """
        Predict celltypes using scANVI.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving scanvi label prediction to adata.obs["{self.result_key}"]')

        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", self.result_key] = self.model.predict(
            adata[adata.obs["_predict_cells"] == "relabel"]
        )
        if self.return_probabilities:
            if f"{self.result_key}_probabilities" not in adata.obs.columns:
                adata.obs[f"{self.result_key}_probabilities"] = pd.Series(dtype="float64")
            adata.obs.loc[
                adata.obs["_predict_cells"] == "relabel",
                f"{self.result_key}_probabilities",
            ] = np.max(
                self.model.predict(adata[adata.obs["_predict_cells"] == "relabel"], soft=True),
                axis=1,
            )

    def compute_umap(self, adata):
        """
        Compute UMAP embedding of integrated data.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obsm[self.umap_key].
        """
        transformer = "rapids" if settings.cuml else None
        sc.pp.neighbors(adata, use_rep=self.embedding_key, transformer=transformer)
        method = "rapids" if settings.cuml else "umap"
        adata.obsm[self.umap_key] = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs).obsm["X_umap"]



from __future__ import annotations

import logging
import os

import joblib
import numpy as np
import pandas as pd
import scanpy as sc
from pynndescent import PyNNDescentTransformer
from scvi.model import SCVI
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class KNN_SCVI(BaseAlgorithm):
    """
    Class to compute KNN classifier after scVI integration.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
        Default is "_batch_annotation".
    labels_key
        Key in obs field of adata for cell-type information.
        Default is "_labels_annotation".
    max_epochs
        Number of epochs scvi is trained.
    result_key
        Key in obs in which celltype annotation results are stored.
        Default is "popv_knn_scvi_prediction".
    embedding_key
        Key in obsm in which latent dimensions are stored.
        Default is "X_scvi_popv".
    umap_key
        Key in obsm in which UMAP embedding of integrated data is stored.
        Default is "X_umap_scvi_popv".
    model_kwargs
        Dictionary to supply non-default values for SCVI model.
        Options at :class:`scvi.model.SCVI`.
        Default is
        {"n_layers": 3, "n_latent": 20, "gene_likelihood": "nb", "use_batch_norm": "none", "use_layer_norm": "both", "encode_covariates": True}.
    classifier_kwargs
        Dictionary to supply non-default values for KNN classifier.
        See :class:`sklearn.neighbors.KNeighborsClassifier`.
        Default is {"weights": "uniform", "n_neighbors": 15}.
    embedding_kwargs
        Dictionary to supply non-default values for UMAP embedding.
        See :func:`scanpy.tl.umap`. Default is {"min_dist": 0.1}.
    train_kwargs
        Dictionary to supply non-default values for training scVI.
        Options at :meth:`scvi.model.SCVI.train`.
        Default is
        {"max_epochs": 20, "batch_size": 512, "accelerator": settings.accelerator, "plan_kwargs": {"n_epochs_kl_warmup": 20}}.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        save_folder: str | None = None,
        result_key: str | None = "popv_knn_on_scvi_prediction",
        embedding_key: str | None = "X_scvi_popv",
        umap_key: str | None = "X_umap_scvi_popv",
        model_kwargs: dict | None = None,
        classifier_dict: dict | None = None,
        embedding_kwargs: dict | None = None,
        train_kwargs: dict | None = None,
    ) -> None:
        super().__init__(
            batch_key=batch_key,
            labels_key=labels_key,
            result_key=result_key,
            embedding_key=embedding_key,
            umap_key=umap_key,
        )
        if embedding_kwargs is None:
            embedding_kwargs = {}
        if classifier_dict is None:
            classifier_dict = {}
        if model_kwargs is None:
            model_kwargs = {}
        if train_kwargs is None:
            train_kwargs = {}
        self.save_folder = save_folder

        self.model_kwargs = {
            "n_layers": 3,
            "n_latent": 20,
            "gene_likelihood": "nb",
            "use_batch_norm": "none",
            "use_layer_norm": "both",
            "encode_covariates": True,
        }

        if model_kwargs is not None:
            self.model_kwargs.update(model_kwargs)

        self.classifier_dict = {"weights": "uniform", "n_neighbors": 15}
        if classifier_dict is not None:
            self.classifier_dict.update(classifier_dict)

        self.train_kwargs = {
            "max_epochs": 20,
            "batch_size": 512,
            "accelerator": settings.accelerator,
            "plan_kwargs": {"n_epochs_kl_warmup": 20},
        }
        self.train_kwargs.update(train_kwargs)
        self.max_epochs = train_kwargs.get("max_epochs", None)

        self.embedding_kwargs = {"min_dist": 0.3}
        if embedding_kwargs is not None:
            self.embedding_kwargs.update(embedding_kwargs)

    def compute_integration(self, adata):
        """
        Compute scVI integration.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obsm[self.embedding_key].
        """
        logging.info("Integrating data with scvi")
        if not adata.uns["_pretrained_scvi_path"]:
            SCVI.setup_anndata(
                adata,
                batch_key=self.batch_key,
                labels_key=self.labels_key,
                layer="scvi_counts",
            )
            model = SCVI(adata, **self.model_kwargs)
            logging.info("Training scvi offline.")
        else:
            query = adata[adata.obs["_predict_cells"] == "relabel"].copy()
            model = SCVI.load_query_data(query, adata.uns["_pretrained_scvi_path"])
            logging.info("Training scvi online.")

        if adata.uns["_prediction_mode"] == "fast":
            self.train_kwargs["max_epochs"] = 1
            model.train(**self.train_kwargs)
        else:
            if self.max_epochs is None:
                self.max_epochs = min(round((20000 / adata.n_obs) * 200), 200)
            model.train(**self.train_kwargs)

            if adata.uns["_save_path_trained_models"] and adata.uns["_prediction_mode"] == "retrain":
                save_path = os.path.join(adata.uns["_save_path_trained_models"], "scvi")
                # Update scvi for scanvi.
                adata.uns["_pretrained_scvi_path"] = save_path
                model.save(
                    save_path,
                    save_anndata=False,
                    overwrite=True,
                )

        latent_representation = model.get_latent_representation()
        relabel_indices = adata.obs["_predict_cells"] == "relabel"
        if self.embedding_key not in adata.obsm:
            # Initialize X_scanvi with the correct shape if it doesn't exist
            adata.obsm[self.embedding_key] = np.zeros((adata.n_obs, latent_representation.shape[1]))
        adata.obsm[self.embedding_key][relabel_indices, :] = latent_representation

    def predict(self, adata):
        """
        Predict celltypes using KNN on scVI embedding.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Saving knn on scvi results to adata.obs["{self.result_key}"]')

        if adata.uns["_prediction_mode"] == "retrain":
            ref_idx = adata.obs["_labelled_train_indices"]

            train_X = adata[ref_idx].obsm[self.embedding_key].copy()
            train_Y = adata.obs.loc[ref_idx, self.labels_key].cat.codes.to_numpy()
            knn = make_pipeline(
                PyNNDescentTransformer(
                    n_neighbors=self.classifier_dict["n_neighbors"],
                    n_jobs=settings.n_jobs,
                ),
                KNeighborsClassifier(metric="precomputed", weights=self.classifier_dict["weights"]),
            )
            knn.fit(train_X, train_Y)
            joblib.dump(
                knn,
                open(
                    os.path.join(
                        adata.uns["_save_path_trained_models"],
                        "scvi_knn_classifier.joblib",
                    ),
                    "wb",
                ),
            )
        else:
            knn = joblib.load(
                open(
                    os.path.join(
                        adata.uns["_save_path_trained_models"],
                        "scvi_knn_classifier.joblib",
                    ),
                    "rb",
                )
            )

        # save_results
        embedding = adata[adata.obs["_predict_cells"] == "relabel"].obsm[self.embedding_key]
        knn_pred = knn.predict(embedding)
        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[adata.obs["_predict_cells"] == "relabel", self.result_key] = adata.uns["label_categories"][
            knn_pred
        ]
        if self.return_probabilities:
            if f"{self.result_key}_probabilities" not in adata.obs.columns:
                adata.obs[f"{self.result_key}_probabilities"] = pd.Series(dtype="float64")
            adata.obs.loc[
                adata.obs["_predict_cells"] == "relabel",
                f"{self.result_key}_probabilities",
            ] = np.max(knn.predict_proba(embedding), axis=1)

    def compute_umap(self, adata):
        """
        Compute UMAP embedding of scVI results.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obsm[self.umap_key].
        """
        if self.compute_umap_embedding:
            logging.info(f'Saving UMAP of scvi results to adata.obs["{self.umap_key}"]')

            transformer = "rapids" if settings.cuml else None
            sc.pp.neighbors(adata, use_rep=self.embedding_key, transformer=transformer)
            method = "rapids" if settings.cuml else "umap"
            adata.obsm[self.umap_key] = sc.tl.umap(adata, copy=True, method=method, **self.embedding_kwargs).obsm[
                "X_umap"
            ]



template = """\
---
{card_data}
---

Popular Vote (popV) model for automated cell type annotation of single-cell RNA-seq data. We provide here pretrained models
for plug-in use in your own analysis.
Follow our [tutorial](https://github.com/YosefLab/popV/blob/main/tabula_sapiens_tutorial.ipynb) to learn how to use the model for cell type annotation.

# Model description

{description}

**Link to CELLxGENE**:
Link to the [data]({cellxgene_url}) in the CELLxGENE browser for interactive exploration of the data and download of the source data.

**Training Code URL**:
{training_code_url}

# Metrics

We provide here accuracies for each of the experts and the ensemble model. The validation set accuracies are
computed on a 10% random subset of the data that was not used for training.

{validation_accuracies}

The train accuracies are computed on the training data.

{train_accuracies}

</details>
\n
# References

{references}
"""



from __future__ import annotations

from abc import abstractmethod

from popv import settings


class BaseAlgorithm:
    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        seen_result_key: str | None = None,
        result_key: str | None = None,
        embedding_key: str | None = None,
        umap_key: str | None = None,
    ) -> None:
        """
        Class to define base algorithm for celltype annotation.

        Parameters
        ----------
        batch_key
            Key in obs field of adata for batch information.
            Default is "_batch_annotation".
        labels_key
            Key in obs field of adata for cell-type information.
            Default is "_labels_annotation".
        seen_result_key
            Key in obs in which celltype predictions are stored. Defaults to result_key.
        result_key
            Key in obs in which celltype predictions are stored.
        embedding_key
            Key in obsm in which integrated embedding is stored.
        umap_key
            Key in obsm in which UMAP embedding of integrated data is stored.
        """
        self.batch_key = batch_key
        self.labels_key = labels_key
        if seen_result_key is None:
            self.seen_result_key = result_key
        else:
            self.seen_result_key = seen_result_key
        self.result_key = result_key
        self.embedding_key = embedding_key
        self.umap_key = umap_key
        self.enable_cuml = settings.cuml
        self.return_probabilities = settings.return_probabilities
        self.compute_umap_embedding = settings.compute_umap_embedding

    @abstractmethod
    def compute_integration(self, adata):
        """Compute integration of adata inplace."""

    @abstractmethod
    def predict(self, adata):
        """Predicts cell type of adata inplace."""

    @abstractmethod
    def compute_umap(self, adata):
        """Compute UMAP embedding of adata inplace."""
